# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pydantic import BaseModel, Field, model_validator
from pydantic.json_schema import GenerateJsonSchema
from pydantic_core import core_schema
from datetime import datetime
from typing import Any, Dict, Union, Optional, List, Literal
import os
import json
import hashlib
from cosmos_rl.utils.modelscope import update_config_if_modelscope


def config_hash(config: BaseModel) -> str:
    """
    Compute the hash of a config object
    """
    if isinstance(config, BaseModel):
        return hashlib.md5(json.dumps(config.model_dump()).encode()).hexdigest()
    else:
        return "unhashable"


class CustomJsonSchemaGenerator(GenerateJsonSchema):
    def generate(
        self, schema: core_schema.CoreSchema, mode="serialization"
    ) -> dict[str, Any]:
        json_schema = super().generate(schema, mode)

        if "properties" in json_schema:
            properties = json_schema["properties"]
            filtered_properties = {
                k: v
                for k, v in properties.items()
                if not (isinstance(v, dict) and v.get("hide_in_doc") is True)
            }
            json_schema["properties"] = filtered_properties

        # Remove 'hide_in_doc' from all the sub-models
        if "$defs" in json_schema:
            defs = json_schema["$defs"]
            for model_def in defs:
                filtered_sub_properties = {
                    k: v
                    for k, v in defs[model_def].get("properties", {}).items()
                    if not (isinstance(v, dict) and v.get("hide_in_doc") is True)
                }
                json_schema["$defs"][model_def]["properties"] = filtered_sub_properties

        return json_schema


class DatasetConfig(BaseModel):
    name: str = Field(
        default="",
        description="Huggingface dataset name or local path to parquet file",
    )

    subset: Optional[str] = Field(
        default="",
        description="Dataset subset if exists",
    )

    revision: Optional[str] = Field(
        default="",
        description={
            "help": "Dataset git revision if exist, can be a branch name, a tag, or a commit hash."
        },
    )

    split: Union[str, List[str]] = Field(
        default="",
        description="A list of dataset splits to train",
    )

    test_size: Optional[Union[float, int]] = Field(
        default=None,
        description="Size of the test set. If float, it is the ratio (between 0.0 and 1.0) of the dataset; if int, it is the absolute size of the test set.",
    )

    @model_validator(mode="after")
    def check_params_value(self):
        if isinstance(self.split, str):
            self.split = [self.split]
        return self


class SFTDataConfig(BaseModel):
    type: Literal["sft"]

    dataset: DatasetConfig = Field(
        default_factory=DatasetConfig,
        description="Dataset configuration for SFT training. It includes dataset name, subset, revision, train split, and test split.",
    )

    mini_batch: int = Field(
        default=2,
        description="mini-batch size for training.",
    )

    dataloader_shuffle: bool = Field(
        default=False,
        description="Shuffle the dataloader. If False, the dataloader will be used in the order it is loaded.",
    )
    enable_dataset_cache: bool = Field(
        default=False,
        description="Enable dataset cache process results, maybe accelerate the dataset loading",
    )
    dataloader_num_workers: int = Field(
        default=0, description="Number of subprocess to use for data loading"
    )
    dataloader_prefetch_factor: Optional[int] = Field(
        default=None,
        description="Number of batches loaded in advance by each worker.",
    )
    conversation_column_name: str = Field(
        default="conversations",  # "conversation",
        description="Column name for formated conversation json",
    )
    system_prompt: str = Field(
        default="",
        description="System prompt for the model, which will be prepended to the prompt",
    )

    @model_validator(mode="after")
    def check_params_value(self):
        if self.dataloader_num_workers <= 0:
            self.dataloader_prefetch_factor = None
            self.dataloader_num_workers = 0
        return self


class CheckpointConfig(BaseModel):
    enable_checkpoint: bool = Field(
        default=False,
        description="Enable checkpointing for training. If set to False, no checkpoint will be saved.",
    )

    save_freq: int = Field(
        default=20, description="Checkpoint save frequency for training steps"
    )
    save_mode: str = Field(
        default="async",
        description="Checkpoint save mode for training steps",
        choices=["async", "sync"],
    )
    max_keep: int = Field(
        default=5,
        description="Maximum number of checkpoints to keep. If set to -1, all checkpoints will be kept.",
    )
    export_safetensors: bool = Field(
        default=True,
        description="Whether to export a safetensors weight for huggingface usage, include related config files.",
    )
    upload_hf: bool = Field(
        default=False,
        description="Whether to upload the safetensors weight to huggingface.",
    )
    hf_repo_name: str = Field(
        default="Comos-Reason1",
        description="The huggingface repo name to upload the safetensors weight.",
    )
    upload_s3: Union[bool, str] = Field(
        default=False,
        description="Whether to upload the checkpoint and safetensors to S3. Default to False, set `final` will upload the final checkpoint, `all` will upload all checkpoints.",
    )
    s3_bucket: Optional[str] = Field(
        default=None,
        description="The S3 bucket name to upload the checkpoint and safetensors weight.",
    )
    s3_prefix: str = Field(
        default="outputs",
        description="The S3 prefix to upload the checkpoint and safetensors weight.",
    )

    @model_validator(mode="after")
    def check_params_value(self):
        if self.upload_s3:
            if self.upload_s3 not in ["final", "all"]:
                raise ValueError(
                    "upload_s3 must be one of ['final', 'all'] or False, got {}".format(
                        self.upload_s3
                    )
                )
            if self.s3_bucket is None:
                raise ValueError(
                    "s3_bucket must be specified when upload_s3 is True, got None"
                )
        if self.save_mode not in ["async", "sync"]:
            raise ValueError(
                f"Invalid save_mode: {self.save_mode}. Must be one of ['async', 'sync']"
            )
        if self.save_freq <= 0:
            raise ValueError(f"save_freq must be greater than 0, got {self.save_freq}")
        return self


class OverlongRewardConfig(BaseModel):
    enable_overlong_penalty: bool = Field(
        default=False,
        description="Enable overlong penalty for the model. If set to True, the output will be penalized for responses that are too long.",
    )
    buffer_length: int = Field(
        default=4096,
        description="Length of the buffer for overlong penalty. If the response length exceeds this value, the output will be penalized.",
    )
    penalty_factor: float = Field(
        default=1.0,
        description="Penalty factor for overlong penalty. The penalty increases linearly with the length of the response exceeding the buffer length from 0 to the penalty_factor.",
    )


class GrpoConfig(BaseModel):
    type: Literal["grpo"]
    variant: str = Field(
        default="grpo",
        description="Variant of the GRPO, currently support `grpo`, and `dapo`",
        choices=["grpo", "dapo"],
    )

    dataset: DatasetConfig = Field(
        default_factory=DatasetConfig,
        description="Dataset configuration for GRPO training. It includes dataset name, subset, revision, train split, test split and test size.",
    )

    dataloader_shuffle: bool = Field(
        default=True,
        description="Shuffle the dataloader. If False, the dataloader will be used in the order it is loaded.",
    )
    enable_dataset_cache: bool = Field(
        default=False,
        description="Enable dataset cache process results, maybe accelerate the dataset loading",
    )
    dataloader_num_workers: int = Field(
        default=0, description="Number of subprocess to use for data loading"
    )
    dataloader_prefetch_factor: Optional[int] = Field(
        default=None,
        description="Number of batches loaded in advance by each worker.",
    )
    prompt_column_name: str = Field(
        default="",
        description="Column name for prompt",
    )
    response_column_name: str = Field(
        default="",
        description="Column name for response/reference answer",
    )
    reward_function: Union[str, List[str], Dict[str, float]] = Field(
        default_factory=lambda: ["single_choice"],
        description="Reward functions for the model. Currently support `single_choice`, `boxed_math`, and `format`. You can add weight to each reward function by passing a dict, e.g., {'single_choice': 0.9, 'format': 0.1}",
    )
    temperature: float = Field(
        default=1.0,
        description="Temperature for sampling. The higher the temperature, the more random the completions.",
    )

    epsilon_low: float = Field(
        default=0.2,
        description="Epsilon value for clipping.",
    )

    epsilon_high: float = Field(
        default=0.2,
        description="Upper-bound epsilon value for clipping. If not specified, it defaults to the same value as the "
        "lower-bound specified in argument `epsilon`. Paper DAPO recommends `0.28`.",
    )

    lower_bound_ratio: float = Field(
        default=3.0,
        description="Lower-bound ratio for dual-clip.",
    )

    loss_type: str = Field(
        default="token-mean",
        description="The type of loss to use for GRPO training.",
        choices=["token-mean", "seq-mean-token-sum", "seq-mean-token-mean"],
    )

    unbiased_loss_max_tokens: Optional[int] = Field(
        default=None,
        description="Maximum number of tokens to use for unbiased loss introduced in Dr.GRPO. If set to None, will not use unbiased loss."
        "Only available when `loss_type` is `seq-mean-token-mean`",
    )

    unbiased_advantage: bool = Field(
        default=False,
        description="Whether to divide the advantage by the standard deviation of rewards.",
    )

    overlong_reward: OverlongRewardConfig = Field(
        default_factory=OverlongRewardConfig,
        description="Configuration for overlong reward penalty. If enabled, the output will be penalized for responses that are too long.",
    )

    kl_beta: float = Field(
        default=0.0,
        description="KL coefficient. If `0.0`, the reference model is not loaded, reducing memory usage and improving "
        "training speed, but may be numerically unstable for long training runs.",
    )

    aipo_rho: Optional[float] = Field(
        default=None,
        description="Rho value for AIPO (Asynchronous Importance weighted Policy Optimization). The clipping constant of the importance sampling ratio, suggest [2,10]. "
        "reference: https://arxiv.org/pdf/2505.24034",
    )

    mu_iterations: int = Field(
        default=1,
        description="Number of iterations per batch (denoted as μ in the algorithm).",
    )

    mini_batch: int = Field(
        default=2,
        description="mini-batch size for GRPO training.",
    )

    allowed_outdated_steps: int = Field(
        default=4,
        description="Allowed outdated-async steps for rollout engine. "
        "If the number of left pending rollouts is larger than the `allowed_outdated_steps * n_policy_replicas * train_batch_per_replica`, "
        "then rollout engine traffic will be throttled. ",
    )

    min_filter_prefix_tokens: Optional[int] = Field(
        default=None,
        description="Minimum number of tokens to filter the prefix tokens for the rollouts inside the same group. "
        "If the number of tokens is larger than the `min_filter_prefix_tokens`, the rollouts with the same prefix but different rewards will be filtered out in loss calculation.",
    )

    @model_validator(mode="after")
    def check_params_value(self):
        assert self.variant in [
            "grpo",
            "dapo",
        ], "variant must be one of ['grpo', 'dapo']"
        if self.dataloader_num_workers <= 0:
            self.dataloader_prefetch_factor = None
            self.dataloader_num_workers = 0
        if isinstance(self.reward_function, str):
            self.reward_function = {self.reward_function: 1.0}
        elif isinstance(self.reward_function, list):
            self.reward_function = {k: 1.0 for k in self.reward_function}
        assert (
            len(self.reward_function) > 0
        ), "reward_function must be a dict of reward functions"
        return self


class SubProfilerConfig(BaseModel):
    do_profile: bool = Field(
        default=False, description="Whether to profile, only used in runtime."
    )
    active_steps: int = Field(default=1, description="Number of active steps")
    warmup_steps: int = Field(default=1, description="Number of warmup steps")
    wait_steps: int = Field(default=1, description="Number of wait steps")
    rank_filter: List[int] = Field(default_factory=list, description="Rank filter")
    record_shape: bool = Field(default=False, description="Whether to record shape")
    profile_memory: bool = Field(default=False, description="Whether to profile memory")
    with_stack: bool = Field(default=False, description="Whether to profile stack")
    with_modules: bool = Field(default=False, description="Whether to profile modules")


class ProfilerConfig(BaseModel):
    enable_profiler: bool = Field(
        default=False,
        description="Enable profiler for training",
    )
    enable_nsys: bool = Field(
        default=False,
        description="Enable nsys for training",
    )
    sub_profiler_config: SubProfilerConfig = Field(
        default_factory=SubProfilerConfig, description="Sub profiler config"
    )


class FP8Config(BaseModel):
    enable_fp8: bool = Field(default=False, description="Whether to enable fp8.")
    fp8_recipe: str = Field(
        default="dynamic_scaling",
        description="Recipe for weight scale calculation.",
        choices=["dynamic_scaling", "delayed_scaling"],
    )
    quant_recipe: str = Field(
        default="rowwise",
        description="Quantization strategy for weight.",
        choices=["rowwise", "tensorwise"],
    )


class TrainingConfig(BaseModel):
    train_policy: Union[SFTDataConfig, GrpoConfig] = Field(
        discriminator="type", default=GrpoConfig(type="grpo")
    )
    fp8: FP8Config = Field(default_factory=FP8Config)
    ckpt: CheckpointConfig = Field(default_factory=CheckpointConfig)
    resume: Union[bool, str] = Field(
        default=False,
        description="Resume training from a checkpoint. If True, will resume from the latest checkpoint of the `output_dir`. If a string, will resume from the specified checkpoint path.",
    )
    epoch: int = Field(default=1, description="Number of epochs for training")
    output_dir: str = Field(default="./outputs", description="Output directory")
    timestamp: str = Field(
        default="",
        description="Timestamp for the output directory and wandb ID, if not set, will be generated automatically",
    )
    epsilon: float = Field(default=1e-6, description="Epsilon for optimizer")
    optm_name: str = Field(
        default="AdamW",
        description="Optimizer name",
        choices=["AdamW", "Adam"],
    )
    optm_lr: Union[float, List[float]] = Field(
        default=1e-6,
        description="Learning rate for optimizer, can be a float or a list of floats for multiple optimizers",
    )
    optm_impl: Union[str, List[str]] = Field(
        default="fused",
        description="Implementation type for optimizer. More info: https://pytorch.org/docs/stable/optim.html, can be a list of strings for multiple optimizers",
        choices=["fused", "foreach", "for-loop"],
    )
    optm_weight_decay: float = Field(
        default=0.01, description="Weight decay for optimizer"
    )
    optm_betas: tuple[float, float] = Field(
        default=(0.9, 0.999), description="Betas for optimizer"
    )
    optm_warmup_steps: int = Field(default=20, description="Warmup steps for optimizer")
    optm_grad_norm_clip: float = Field(
        default=1.0, description="Gradient norm clip for optimizer"
    )

    # --------- smoke-test helpers ---------
    max_num_steps: Optional[int] = Field(
        default=None,
        description="Optional upper bound on total training steps. If set, training stops when either this step count or the epoch-based limit is reached (whichever comes first). Handy for quick smoke tests.",
    )

    async_tp_enabled: bool = Field(
        default=False, description="Whether to use async tensor parallelism"
    )

    compile: bool = Field(default=True, description="Whether to use torch.compile")

    param_dtype: str = Field(
        default="bfloat16",
        description="The data type for parameters and activations",
        choices=["bfloat16", "float16", "float32"],
    )

    fsdp_reduce_dtype: str = Field(
        default="float32",
        description="The data type for reduction in FSDP",
        choices=["float32"],
    )
    fsdp_offload: bool = Field(
        default=False,
        description="Whether to offload the model to CPU if using FSDP",
    )

    fsdp_reshard_after_forward: str = Field(
        default="default",
        description="Reshard the param after forward pass in FSDP",
        choices=["always", "never", "default"],
    )

    train_batch_per_replica: int = Field(
        default=8,
        description="The batch size for training per iteration in one replica, this is the local batch size for each gradient accumulation step",
    )

    enable_validation: bool = Field(
        default=False,
        description="Enable validation during training.",
    )
    validation_step: int = Field(
        default=20,
        description="Validation frequency during training, in terms of training steps",
    )
    validation_batch_per_replica: int = Field(
        default=24,
        description="The batch size for validation per iteration in one replica.",
    )

    sync_weight_interval: int = Field(
        default=1,
        description="The interval of train step for synchronizing weights between replicas.",
    )

    @model_validator(mode="after")
    def check_params_value(self):
        if self.async_tp_enabled and not self.compile:
            raise ValueError(
                "Async tensor parallelism requires torch.compile to be enabled"
            )
        if self.max_num_steps is not None and self.max_num_steps <= 0:
            raise ValueError("max_num_steps must be positive if specified")
        return self


class ParallelismConfig(BaseModel):
    n_init_replicas: int = Field(
        default=1,
        description="Number of initial replicas to be created",
    )
    tp_size: int = Field(default=2, description="Tensor parallelism size")
    cp_size: int = Field(default=1, description="Context parallelism size")
    dp_shard_size: int = Field(
        default=-1, description="Data Parallelism size in sharded mode"
    )
    pp_size: int = Field(default=1, description="Pipeline parallelism size")
    pp_dynamic_shape: bool = Field(
        default=False, description="Pipeline parallelism dynamic shape"
    )
    pp_micro_batch_size: int = Field(
        default=1,
        description="Pipeline parallelism micro batch size, `n_micro_batch = batch_size / pp_micro_batch_size`, which must be divisible by `pp` stages",
    )
    dp_replicate_size: int = Field(
        default=1,
        description="Data Parallelism size in replica mode. Only configurable in SFT type job, must be 1 in GRPO type job for dynamic scaling support purpose.",
        choices=[1],
    )

    @property
    def world_size(self):
        world_size = os.environ.get("WORLD_SIZE", 1)
        return int(world_size)

    @property
    def local_world_size(self):
        local_world_size = os.environ.get("LOCAL_WORLD_SIZE", 1)
        return int(local_world_size)


class PolicyConfig(BaseModel):
    parallelism: ParallelismConfig = Field(default_factory=ParallelismConfig)
    model_name_or_path: str = Field(
        # default="Qwen/Qwen2.5-3B-Instruct",  #'Qwen/Qwen2.5-VL-7B-Instruct'
        default="Qwen/Qwen2.5-VL-7B-Instruct",
        description="The model name or path, compatible with huggingface model name or local path",
    )
    model_max_length: int = Field(
        default=4096,
        description="The maximum length for training, longer than this will be ignored for training stability",
    )
    model_gradient_checkpointing: bool = Field(
        default=True, description="Whether to use gradient checkpointing"
    )

    @model_validator(mode="after")
    def check_params_value(self):
        assert (
            self.model_name_or_path is not None and self.model_name_or_path != ""
        ), "model_name_or_path is required"
        assert self.parallelism.tp_size > 0, "tp_size must be greater than 0"
        assert self.parallelism.cp_size > 0, "cp_size must be greater than 0"
        assert self.parallelism.pp_size > 0, "pp_size must be greater than 0"
        assert (
            self.parallelism.dp_shard_size >= -1 and self.parallelism.dp_shard_size != 0
        ), "dp_shard_size must be greater than 0 or -1 to be auto-inferred"
        return self


class RolloutParallelismConfig(ParallelismConfig):
    n_init_replicas: int = Field(
        default=1, description="Number of initial replicas to be created"
    )
    tp_size: int = Field(default=2, description="Tensor parallelism size")
    pp_size: int = Field(default=1, description="Pipeline parallelism size")

    # Fields below are that we do not want user to config it.
    dp_replicate_size: int = Field(
        default=1,
        description="Data Parallelism size in replica mode, only 1 is supported for dynamic scaling purpose.",
        choices=[1],
    )
    cp_size: int = Field(default=1, description="Context parallelism size")
    dp_shard_size: int = Field(
        default=-1, description="Data Parallelism size in sharded mode"
    )


class SamplingConfig(BaseModel):
    temperature: float = Field(default=1.0, description="Temperature for sampling.")
    top_p: float = Field(default=1.0, description="Top-p for sampling.")
    top_k: int = Field(default=-1, description="Top-k for sampling.")
    repetition_penalty: float = Field(
        default=1.0, description="Repetition penalty for sampling."
    )
    use_flashinfer: bool = Field(
        default=False, description="Use flashinfer for sampling."
    )


class ValidationConfig(BaseModel):
    dataset: DatasetConfig = Field(
        default_factory=DatasetConfig,
        description="Dataset configuration for validation. It includes dataset name, subset, revision and test split.",
    )

    temperature: float = Field(
        default=0.9, description="Temperature for sampling during validation."
    )
    top_p: float = Field(
        default=1.0, description="Top-p for sampling during validation."
    )
    top_k: int = Field(default=10, description="Top-k for sampling during validation.")
    repetition_penalty: float = Field(
        default=1.0, description="Repetition penalty for sampling during validation."
    )
    n_generation: int = Field(
        default=1,
        description="n parameter same like what in OpenAI chat API for validation.",
    )
    max_response_length: int = Field(
        default=2048,
        description="Max output length of rollout generation during validation.",
    )
    reward_function: Union[str, List[str], Dict[str, float]] = Field(
        default_factory=lambda: ["single_choice"],
        description="Reward functions for the model. Currently support `single_choice`, `boxed_math`, and `format`. You can add weight to each reward function by passing a dict, e.g., {'single_choice': 0.9, 'format': 0.1}",
    )

    @model_validator(mode="after")
    def check_params_value(self):
        if isinstance(self.reward_function, str):
            self.reward_function = {self.reward_function: 1.0}
        elif isinstance(self.reward_function, list):
            self.reward_function = {k: 1.0 for k in self.reward_function}
        assert (
            len(self.reward_function) > 0
        ), "reward_function must be a dict of reward functions"
        return self


class RolloutConfig(BaseModel):
    parallelism: RolloutParallelismConfig = Field(
        default_factory=RolloutParallelismConfig
    )
    enforce_eager: bool = Field(
        default=True, description="Whether to enable eager execution for vLLM."
    )
    include_stop_str_in_output: bool = Field(
        default=False, description="Whether to include stop string in output."
    )
    gpu_memory_utilization: float = Field(
        default=0.8,
        description="GPU memory utilization factor for rollout backend.",
    )
    enable_chunked_prefill: bool = Field(
        default=False, description="Whether to enable chunked prefill for vLLM."
    )
    max_response_length: int = Field(
        default=2048, description="Max output length of rollout generation."
    )
    n_generation: int = Field(
        default=16, description="n parameter same like what in OpenAI chat API."
    )

    batch_size: int = Field(default=1, description="Batch size for rollout.")
    val_batch_size: Optional[int] = Field(
        default=None,
        description="Batch size for rollout generation during validation.",
    )

    quantization: str = Field(
        default="none",
        description="Quantization in vllm rollout generation.",
        choices=["none", "fp8"],
    )

    seed: Optional[int] = Field(default=None, description="random seed for rollout.")

    sampling_config: SamplingConfig = Field(default_factory=SamplingConfig)

    vllm_use_flashinfer: bool = Field(
        default=False, description="Use flashinfer for vllm rollout."
    )

    @model_validator(mode="after")
    def check_params_value(self):
        if isinstance(self.parallelism, dict):
            self.parallelism = RolloutParallelismConfig(**self.parallelism)
        return self


class LoggingConfig(BaseModel):
    logger: List[str] = Field(
        default_factory=list,
        description="List of loggers to use, e.g., ['console', 'wandb']",
    )
    project_name: str = Field(
        default="cosmos_rl",
        description="Wandb project name for logging. If set, the training will be logged to this project.",
    )
    experiment_name: Optional[str] = Field(
        default=None,
        description="A short display name for this run. If not set, will use the `output_dir` as the experiment name.",
    )
    report_mfu: bool = Field(
        default=False,
        description="Whether to report the MFU (Model FLOPs Utilization) to wandb.",
        json_schema_extra={"hide_in_doc": True},
    )

    @model_validator(mode="after")
    def check_params_value(self):
        if self.logger:
            self.logger = [logger.lower() for logger in self.logger]
        return self


class Config(BaseModel):
    train: TrainingConfig = Field(default_factory=TrainingConfig)
    rollout: RolloutConfig = Field(default_factory=RolloutConfig)
    policy: PolicyConfig = Field(default_factory=PolicyConfig)
    logging: LoggingConfig = Field(default_factory=LoggingConfig)
    profiler: ProfilerConfig = Field(default_factory=ProfilerConfig)
    validation: ValidationConfig = Field(default_factory=ValidationConfig)
    redis: str = Field(
        default="",
        description="Redis server address port, format: port",
        json_schema_extra={"hide_in_doc": True},
    )
    eth_ips: str = Field(
        default="",
        description="List of eth ip addresses, format: ip1;ip2;ip3",
        json_schema_extra={"hide_in_doc": True},
    )

    @classmethod
    def from_dict(cls, config_data: dict[str, Any]) -> "Config":
        if "train" in config_data:
            # Set unique timestamp for output directory
            if (
                "timestamp" not in config_data["train"]
                or config_data["train"]["timestamp"] == ""
            ):
                config_data["train"]["timestamp"] = datetime.now().strftime(
                    "%Y%m%d%H%M%S"
                )
                config_data["train"]["output_dir"] = os.path.join(
                    config_data["train"]["output_dir"],
                    config_data["train"]["timestamp"],
                )
        config = cls.model_validate(config_data)
        config = update_config_if_modelscope(config)
        return config

    @model_validator(mode="before")
    def preprocess(cls, data: dict) -> dict:
        # Handle for train_policy type
        if len(data) == 0:
            # empty data, all fields set to default values.
            return data

        if "train_policy" in data["train"]:
            train_policy_data = data["train"]["train_policy"]

            # Determine the type based on characteristic fields
            if any(
                key in train_policy_data
                for key in ["temperature", "epsilon_low", "epsilon_high", "kl_beta"]
            ):
                data["train"]["train_policy"]["type"] = "grpo"
            else:
                data["train"]["train_policy"]["type"] = "sft"
        return data

    @model_validator(mode="after")
    def check_params_value(self):
        if self.policy.parallelism.pp_size > 1:
            assert (
                self.policy.parallelism.pp_micro_batch_size > 0
            ), "pp_micro_batch_size must be greater than 0"
            assert (
                self.train.train_batch_per_replica
                % self.policy.parallelism.pp_micro_batch_size
                == 0
            ), "train_batch must be divisible by pp_micro_batch_size"

            # Here we assume that PP uses `Single-stage per rank` which is true for:
            #   - GPipe
            #   - 1F1B
            # But not correct for those `InterleavedXXX` style schedule
            assert (
                (
                    self.train.train_batch_per_replica
                    // self.policy.parallelism.pp_micro_batch_size
                )
                % self.policy.parallelism.pp_size
                == 0
            ), "train_batch / pp_micro_batch_size must be divisible by pp_size"

        if self.train.train_policy.type == "grpo":
            # Handle for evaludation configuration.
            if isinstance(self.validation.dataset.split, str):
                self.validation.dataset.split = [self.validation.dataset.split]
        return self


COSMOS_CONFIG_SCHEMA = Config.model_json_schema(
    schema_generator=CustomJsonSchemaGenerator
)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Common utilities for the CLI.
"""

import requests
import sys
from typing import Dict, List
from rich.console import Console
from rich.table import Table

from cosmos_rl.utils.network_util import make_request_with_retry
from cosmos_rl.utils.api_suffix import COSMOS_API_STATUS_SUFFIX
from urllib.parse import urljoin

console = Console()


def check_controller_running(base_url: str):
    """
    Check if the controller is running on the given address and port.
    Args:
        address (str): The address of the controller.
        port (str): The port of the controller.
    Returns:
        bool: True if the controller is running, False otherwise.
    """
    url = urljoin(base_url, COSMOS_API_STATUS_SUFFIX)
    is_running = False
    try:
        response = make_request_with_retry(
            requests.get,
            [url],
            max_retries=4,
        )
        if response.status_code == 200:
            is_running = True
        else:
            is_running = False
    except Exception as e:
        console.print(f"Exception when checking controller running: {e}")
        is_running = False
    if not is_running:
        console.print(f"Please check if the controller is running on {base_url}")
        sys.exit(1)
    return is_running


def get_base_url(controller_host: str, controller_port: str):
    base_url = f"http://{controller_host}:{controller_port}"
    if not check_controller_running(base_url):
        console.print(f"Controller is not running on {base_url}.")
        sys.exit(1)
    return base_url


def create_table_for_single_replica(replica: Dict):
    table = Table(
        title=f"Replica : {replica['name']}, Role: {replica['role']}",
        show_lines=True,
    )

    table.add_column("Atom Global Rank")
    table.add_column("Host IP")
    table.add_column("Host Name")
    table.add_column("Trace Path")

    for atom in replica["atoms"]:
        atom_global_rank = atom["global_rank"]
        host_ip = atom["host_ip"]
        host_name = atom["host_name"]
        trace_path = atom["trace_path"]

        table.add_row(
            str(atom_global_rank),
            host_ip,
            host_name,
            trace_path if trace_path else "None",
        )

    return table


def create_table_for_all_replicas(replicas: List[Dict]):
    """
    Create simplified table for all replicas.
    """
    table = Table(
        title="Replicas",
        show_lines=True,
    )
    # One Row for each replica
    table.add_column("Replica")
    table.add_column("Role")
    table.add_column("Arrived")
    table.add_column("Atom Number")
    table.add_column("Weight Step")
    table.add_column("Pending Rollouts")

    for replica in replicas:
        if replica["arrived"]:
            arrived_string = "[green]Yes[/green]"
        else:
            arrived_string = "[red]No[/red]"
        atom_number = len(replica["atoms"])
        if replica["role"] == "POLICY":
            pending_rollouts = replica["pending_rollouts"]
        else:
            pending_rollouts = "N/A"
        table.add_row(
            replica["name"],
            replica["role"],
            arrived_string,
            str(atom_number),
            str(replica["weight_step"]),
            str(pending_rollouts),
        )
    return table



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cosmos_rl.policy.model.gpt import GPT
from cosmos_rl.policy.model.qwen2_5_vl import Qwen2_5_VLConditionalModel
from cosmos_rl.policy.model.qwen3_moe import Qwen3MoE
from cosmos_rl.policy.model.hf_llm import HFLLMModel
from cosmos_rl.policy.model.base import ModelRegistry, BaseModel, WeightMapper

__all__ = [
    "GPT",
    "Qwen2_5_VLConditionalModel",
    "Qwen3MoE",
    "HFLLMModel",
    "BaseModel",
    "WeightMapper",
    "ModelRegistry",
]



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional, Type, Callable
import copy
import uuid
from abc import ABC
from strenum import StrEnum
import msgpack
from cosmos_rl.dispatcher.replica import Replica
from cosmos_rl.dispatcher.protocol import Role
from cosmos_rl.utils.redis_stream import RedisStreamHandler


class CommandType(StrEnum):
    WEIGHT_RESUME = "WEIGHT_RESUME"
    BUILD_MESH = "BUILD_MESH"
    POLICY_TO_POLICY_BROADCAST = "POLICY_TO_POLICY_BROADCAST"
    POLICY_TO_POLICY_UNICAST = "POLICY_TO_POLICY_UNICAST"
    POLICY_TO_ROLLOUT_UNICAST = "POLICY_TO_ROLLOUT_UNICAST"
    ROLLOUT_TO_ROLLOUT_BROADCAST = "ROLLOUT_TO_ROLLOUT_BROADCAST"
    DATA_FETCH = "DATA_FETCH"
    ALL_REDUCE = "ALL_REDUCE"
    STOP = "STOP"
    VALIDATE = "VALIDATE"
    DUMMY = "DUMMY"


class CommandScope:
    GLOBAL = 0
    LOCAL = 1


class Command(ABC):
    uuid_value: str

    def __init__(
        self,
        scope: CommandScope,
        command_type: CommandType,
        uuid_value: Optional[str] = None,
    ):
        self.uuid_value = uuid_value if uuid_value is not None else str(uuid.uuid4())
        self.scope = scope
        self.command_type = command_type

    def _serialize(self) -> Dict:
        dict_v = copy.deepcopy(self.__dict__)
        return dict_v

    def pack(self):
        return msgpack.packb(self.__dict__)

    @classmethod
    def deserialize(cls, dict_v):
        sub_cls = None
        if dict_v["command_type"] == CommandType.WEIGHT_RESUME:
            sub_cls = WeightResumeCommand
        elif dict_v["command_type"] == CommandType.BUILD_MESH:
            sub_cls = BuildMeshCommand
        elif dict_v["command_type"] == CommandType.POLICY_TO_POLICY_BROADCAST:
            sub_cls = PolicyToPolicyBroadcastCommand
        elif dict_v["command_type"] == CommandType.POLICY_TO_POLICY_UNICAST:
            sub_cls = PolicyToPolicyUnicastCommand
        elif dict_v["command_type"] == CommandType.POLICY_TO_ROLLOUT_UNICAST:
            sub_cls = PolicyToRolloutUnicastCommand
        elif dict_v["command_type"] == CommandType.ROLLOUT_TO_ROLLOUT_BROADCAST:
            sub_cls = RolloutToRolloutBroadcastCommand
        elif dict_v["command_type"] == CommandType.DATA_FETCH:
            sub_cls = DataFetchCommand

        if sub_cls is None:
            raise ValueError(f"Unknown command type: {dict_v['command_type']}")
        else:
            return sub_cls.from_dict(dict_v)

    @classmethod
    def depack(cls, byte):
        dict = msgpack.unpackb(byte)
        return cls.deserialize(dict)

    @classmethod
    def from_dict(cls, dict_v: Dict):
        raise NotImplementedError("from_dict is not implemented for Base Command")


class WeightResumeCommand(Command):
    def __init__(self, replica_name: str, **kwargs):
        kwargs["scope"] = CommandScope.LOCAL
        kwargs["command_type"] = CommandType.WEIGHT_RESUME
        super().__init__(**kwargs)
        self.replica_name = replica_name

    replica_name: str

    @classmethod
    def trigger(cls, replica: Replica, redis_handler: RedisStreamHandler):
        assert (
            replica.role == Role.POLICY
        ), "WeightResumeCommand can only be triggered on policy replicas"
        cmd = cls(replica.name)
        redis_handler.publish_command(cmd.pack(), replica.name)
        # initial weight step
        replica.weights_loaded_in_view_of_command = True

    @classmethod
    def from_dict(cls, dict_v: Dict):
        return cls(**dict_v)


class BuildMeshCommand(Command):
    def __init__(self, replica_name_to_rank: Dict[str, int], **kwargs):
        kwargs["scope"] = CommandScope.GLOBAL
        kwargs["command_type"] = CommandType.BUILD_MESH
        super().__init__(**kwargs)
        self.replica_name_to_rank = replica_name_to_rank

    replica_name_to_rank: Dict[str, int]

    @classmethod
    def trigger(cls, replicas: List[Replica], redis_handler: RedisStreamHandler):
        index = 0
        assert all(
            replica.all_atoms_arrived for replica in replicas
        ), "All replicas must have arrived"
        replicas_to_rank = {}
        for replica in replicas:
            replica.status.mesh_rank = index
            replicas_to_rank[replica.name] = index
            index += 1
        cmd = cls(replicas_to_rank)
        for replica in replicas:
            redis_handler.publish_command(cmd.pack(), replica.name)

    @classmethod
    def from_dict(cls, dict_v: Dict):
        return cls(**dict_v)


class PolicyToPolicyBroadcastCommand(Command):
    """
    Only used for policy weight init during initialization. (After `WeightResumeCommand` on `src_replica_name`)
    """

    def __init__(self, src_replica_name: str, dst_replica_names: List[str], **kwargs):
        kwargs["scope"] = CommandScope.GLOBAL
        kwargs["command_type"] = CommandType.POLICY_TO_POLICY_BROADCAST
        super().__init__(**kwargs)
        self.src_replica_name = src_replica_name
        self.dst_replica_names = dst_replica_names

    src_replica_name: str
    dst_replica_names: List[str]

    @classmethod
    def trigger(
        cls,
        src_replica: Replica,
        dst_replicas: List[Replica],
        redis_handler: RedisStreamHandler,
    ):
        # dst_replicas will contains the src_replica
        cmd = cls(src_replica.name, [replica.name for replica in dst_replicas])
        for replica in dst_replicas:
            redis_handler.publish_command(cmd.pack(), replica.name)
            replica.weights_loaded_in_view_of_command = True

    @classmethod
    def from_dict(cls, dict_v: Dict):
        return cls(**dict_v)


class PolicyToPolicyUnicastCommand(Command):
    """
    Used for policy dynamic scaling.
    """

    def __init__(self, src_replica_name: str, dst_replica_name: str, **kwargs):
        kwargs["scope"] = CommandScope.LOCAL
        kwargs["command_type"] = CommandType.POLICY_TO_POLICY_UNICAST
        super().__init__(**kwargs)
        self.src_replica_name = src_replica_name
        self.dst_replica_name = dst_replica_name

    src_replica_name: str
    dst_replica_name: str

    @classmethod
    def trigger(
        cls,
        src_replica: Replica,
        dst_replica: Replica,
        redis_handler: RedisStreamHandler,
    ):
        cmd = cls(src_replica.name, dst_replica.name)
        redis_handler.publish_command(cmd.pack(), src_replica.name)
        redis_handler.publish_command(cmd.pack(), dst_replica.name)
        dst_replica.weights_loaded_in_view_of_command = True

    @classmethod
    def from_dict(cls, dict_v: Dict):
        return cls(**dict_v)


class PolicyToRolloutUnicastCommand(Command):
    """
    Used for:
        - weight updating of rollout for on-policy training.
        - weight initialization of first rollout replica.
    """

    _do_weight_sync_check_flag: bool = True

    def __init__(
        self,
        src_replica_name: str,
        dst_replica_name: str,
        src_replica_size: int,
        dst_replica_size: int,
        do_weight_sync_check: bool = False,
        weight_step: Optional[int] = None,
        total_steps: Optional[int] = None,
        **kwargs,
    ):
        kwargs["scope"] = CommandScope.LOCAL
        kwargs["command_type"] = CommandType.POLICY_TO_ROLLOUT_UNICAST
        super().__init__(**kwargs)
        self.src_replica_name = src_replica_name
        self.dst_replica_name = dst_replica_name
        self.src_replica_size = src_replica_size
        self.dst_replica_size = dst_replica_size
        self.do_weight_sync_check = do_weight_sync_check
        self.weight_step = weight_step
        self.total_steps = total_steps

    src_replica_name: str
    dst_replica_name: str
    src_replica_size: int
    dst_replica_size: int
    do_weight_sync_check: bool
    weight_step: Optional[int]
    total_steps: Optional[int]

    @classmethod
    def trigger(
        cls,
        src_replica: Replica,  # Policy Replica
        dst_replica: Replica,  # Rollout Replica
        src_replica_size: int,
        dst_replica_size: int,
        weight_step: Optional[int],
        total_steps: Optional[int],
        redis_handler: RedisStreamHandler,
    ):
        cmd = cls(
            src_replica.name,
            dst_replica.name,
            src_replica_size,
            dst_replica_size,
            cls._do_weight_sync_check_flag,
            weight_step,
            total_steps,
        )
        redis_handler.publish_command(cmd.pack(), src_replica.name)
        redis_handler.publish_command(cmd.pack(), dst_replica.name)
        dst_replica.weights_loaded_in_view_of_command = True

        if cls._do_weight_sync_check_flag:
            cls._do_weight_sync_check_flag = False

    @classmethod
    def from_dict(cls, dict_v: Dict):
        return cls(**dict_v)


class RolloutToRolloutBroadcastCommand(Command):
    """
    Used for rollout weight update.(After `PolicyToRolloutUnicastCommand` on `src_replica_name`)
    """

    def __init__(
        self,
        src_replica_name: str,
        dst_replica_names: List[str],
        weight_step: Optional[int],
        total_steps: Optional[int],
        **kwargs,
    ):
        kwargs["scope"] = CommandScope.GLOBAL
        kwargs["command_type"] = CommandType.ROLLOUT_TO_ROLLOUT_BROADCAST
        super().__init__(**kwargs)
        self.src_replica_name = src_replica_name
        self.dst_replica_names = dst_replica_names
        self.weight_step = weight_step
        self.total_steps = total_steps

    src_replica_name: str
    dst_replica_names: List[str]
    weight_step: Optional[int]
    total_steps: Optional[int]

    @classmethod
    def trigger(
        cls,
        src_replica: Replica,
        dst_replicas: List[Replica],
        weight_step: Optional[int],
        total_steps: Optional[int],
        redis_handler: RedisStreamHandler,
    ):
        # dst_replicas will contains the src_replica
        if not src_replica.in_mesh:
            return
        cmd = cls(
            src_replica.name,
            [replica.name for replica in dst_replicas],
            weight_step,
            total_steps,
        )
        for replica in dst_replicas:
            if not replica.in_mesh:
                continue
            redis_handler.publish_command(cmd.pack(), replica.name)
            replica.weights_loaded_in_view_of_command = True

    def replica_should_stop(self):
        if self.weight_step is not None and self.total_steps is not None:
            return self.weight_step >= self.total_steps
        return False

    @classmethod
    def from_dict(cls, dict_v: Dict):
        return cls(**dict_v)


class DataFetchCommand(Command):
    """
    Used to fetch data from the controller.
    items_count: int,  Number of items to fetch.
    replica_name: str, Name of the replica to fetch data.
    """

    def __init__(
        self,
        replica_name: str,
        items_count: int,
        global_step: int,
        total_steps: int,
        remain_samples_num: int,
        # For profiling
        do_profile: Optional[bool] = None,
        active_steps: Optional[int] = None,
        rank_filter: Optional[List[int]] = None,
        record_shape: Optional[bool] = None,
        profile_memory: Optional[bool] = None,
        with_stack: Optional[bool] = None,
        with_modules: Optional[bool] = None,
        **kwargs,
    ):
        kwargs["scope"] = CommandScope.LOCAL
        kwargs["command_type"] = CommandType.DATA_FETCH
        super().__init__(**kwargs)
        self.replica_name = replica_name
        self.items_count = items_count
        self.global_step = global_step
        self.total_steps = total_steps
        self.remain_samples_num = remain_samples_num

        # Profling config
        self.do_profile = do_profile
        self.active_steps = active_steps
        self.rank_filter = rank_filter
        self.record_shape = record_shape
        self.profile_memory = profile_memory
        self.with_stack = with_stack
        self.with_modules = with_modules

    replica_name: str
    items_count: int
    global_step: Optional[int]
    total_steps: Optional[int]
    remain_samples_num: int

    do_profile: bool
    active_steps: int
    rank_filter: List[int]
    record_shape: bool
    profile_memory: bool
    with_stack: bool
    with_modules: bool

    @classmethod
    def trigger(
        cls,
        replica: Replica,
        items_count: int,
        global_step: Optional[int],
        total_steps: Optional[int],
        remain_samples_num: int,
        redis_handler: RedisStreamHandler,
    ):
        cmd = cls(
            replica.name,
            items_count,
            global_step,
            total_steps,
            remain_samples_num,
            replica.sub_profiler_config.do_profile,
            replica.sub_profiler_config.active_steps,
            replica.sub_profiler_config.rank_filter,
            replica.sub_profiler_config.record_shape,
            replica.sub_profiler_config.profile_memory,
            replica.sub_profiler_config.with_stack,
            replica.sub_profiler_config.with_modules,
        )
        redis_handler.publish_command(cmd.pack(), replica.name)

    def replica_should_stop(self):
        if self.global_step is not None and self.total_steps is not None:
            return self.global_step >= self.total_steps
        return False

    @classmethod
    def from_dict(cls, dict_v: Dict):
        return cls(**dict_v)


class CommandRegistry:
    def __init__(self):
        self.registry: Dict[Type[Command], Callable] = {}

    def register(self, key: Type[Command], func: Callable):
        if key not in self.registry:
            self.registry[key] = func
        else:
            raise ValueError(f"Command handler for {key} already registered")

    def get_command_handler(self, key: Type[Command]) -> Optional[Callable]:
        return self.registry.get(key, None)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import torch
import torch.distributed as dist
import subprocess

# For testing, set the timeout to 60000ms
os.environ["COSMOS_NCCL_TIMEOUT_MS"] = "60000"
from cosmos_rl.utils.distributed import broadcast_object_cpu, nccl_timeout_watchdog
from cosmos_rl.utils.pynccl import (
    create_nccl_uid,
    create_nccl_comm,
    nccl_allreduce,
    nccl_abort,
)


def routine(N: int, device: torch.device, rank: int, world_size: int):
    nccl_uid = broadcast_object_cpu(create_nccl_uid() if rank == 0 else None)
    stream = torch.cuda.Stream()
    with torch.cuda.stream(stream):
        comm = create_nccl_comm(nccl_uid, rank, world_size)
        try:
            with nccl_timeout_watchdog(wait_stream=True):
                send_buffer = torch.arange(N, dtype=torch.float32, device=device)
                recv_buffer = torch.zeros(N, dtype=torch.float32, device=device)

                if rank != 3:
                    # Simulate a failed allreduce on rank 3
                    try:
                        nccl_allreduce(send_buffer, recv_buffer, 0, comm)
                        print(f"[RANK {rank}] arrived here")
                    except Exception as e:
                        print(f"[RANK {rank}] error in nccl_allreduce: {e}")
                else:
                    nccl_abort(comm)
        except Exception as e:
            print(f"[RANK {rank}] error in nccl_timeout_watchdog: {e}")

        from cosmos_rl.utils import pynccl as _pynccl

        print("-" * 80)
        print(
            f"[rank {rank}] comm_store keys after exit watchdog: {list(_pynccl._COMM_REGISTRY._store.keys())}"
        )
        print(f"[rank {rank}] task queue size: {_pynccl._task_q.qsize()}")
        print("-" * 80)

        if rank < 2:
            # Test that the new group still works after the previous nccl failure
            print(f"[rank {rank}] creating 2nd comm")
            world_size = 2
            group_ranks = [0, 1]
            new_group = dist.new_group(ranks=group_ranks, backend="cuda:nccl,cpu:gloo")

            nccl_uid = broadcast_object_cpu(
                create_nccl_uid() if rank == 0 else None, group=new_group
            )
            comm = create_nccl_comm(nccl_uid, rank, world_size)
            send_tensor = torch.arange(N, dtype=torch.float32, device=device)
            recv_tensor = torch.zeros(N, dtype=torch.float32, device=device)
            nccl_allreduce(send_tensor, recv_tensor, 0, comm)
            torch.testing.assert_close(
                recv_tensor,
                torch.arange(N, dtype=torch.float32, device=device) * world_size,
            )
            nccl_abort(comm)

        print(f"[rank {rank}] finished the test")


def main():
    local_rank = int(os.environ["LOCAL_RANK"])
    torch.cuda.set_device(local_rank)
    device = torch.device(f"cuda:{local_rank}")
    dist.init_process_group(backend="cuda:nccl,cpu:gloo")

    rank = dist.get_rank()
    world_size = dist.get_world_size()

    ###
    # Global timeout: 60s
    # 1. We initialize comm and launch allreduce on all ranks except rank 3
    # 2. Rank 3 will sleep for 100s
    #     By design, the allreduce could trigger a cpu waiting timeout on all ranks except rank 3
    # 3. Rank 0 and 1 will continue to run and complete the allreduce
    routine(N=8, device=device, rank=rank, world_size=world_size)

    print(f"[RANK {rank}] Running 2nd routine\n")
    from importlib.metadata import version

    nccl_v = version("nvidia-nccl-cu12")
    if nccl_v >= "2.26.2":
        routine(N=512 * 4096 * 512, device=device, rank=rank, world_size=world_size)

    # Global barrier to ensure all ranks finished routines
    dist.barrier()

    # Wait background worker to drain queue and communicator store
    from cosmos_rl.utils import pynccl as _pynccl

    def _wait_for_cleanup(timeout_s: float = 10.0, interval: float = 0.02) -> bool:
        import time

        start = time.time()
        while time.time() - start < timeout_s:
            if _pynccl._task_q.qsize() == 0 and len(_pynccl._COMM_REGISTRY._store) == 0:
                return True
            time.sleep(interval)
        return False

    clean_ok = _wait_for_cleanup()
    if dist.get_rank() == 0:
        print("=" * 80)
        pending_tasks = list(_pynccl._task_q.queue)
        print(f"[test] remaining task count: {len(pending_tasks)}")
        if pending_tasks:
            print("[test] pending tasks:")
            for t in pending_tasks:
                print("   ", t)

        print(
            f"[test] remaining comm_store keys: {list(_pynccl._COMM_REGISTRY._store.keys())}"
        )

        if clean_ok:
            print(
                "[test] cleanup succeeded – task queue and comm store are empty. All tests passed."
            )
        else:
            raise AssertionError(
                "[test] cleanup timed out – background worker did not drain queues in time."
            )

    dist.destroy_process_group()


if __name__ == "__main__":
    if os.environ.get("RECURSIVE_ENTRYPOINT") is None:
        n_gpu = torch.cuda.device_count()
        command = [
            "torchrun",
            "--nnodes",
            "1",
            "--nproc_per_node",
            str(n_gpu),
            "--rdzv_backend",
            "c10d",
            "--rdzv_endpoint",
            "localhost:12345",
            os.path.abspath(__file__),
        ]
        env = os.environ.copy()
        env["RECURSIVE_ENTRYPOINT"] = "1"
        subprocess.run(command, env=env)
    else:
        main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Pass `--stream` (or `--verbose`) when running this script to print all training logs in real time.
import os
import sys
import subprocess
import tempfile
import toml
import re
import threading
import time


from cosmos_rl.utils import util
import torch
import signal

try:
    import psutil
except ImportError:
    psutil = None


def _calc_world_size(parallel_cfg: dict, is_rollout: bool = False) -> int:
    tp = int(parallel_cfg.get("tp_size", 1))
    pp = int(parallel_cfg.get("pp_size", 1))
    if is_rollout:
        return tp * pp or 1
    dp_shard = int(parallel_cfg.get("dp_shard_size", 1))
    if dp_shard == -1:
        dp_shard = 1  # fallback when auto infer disabled in smoke test
    dp_rep = int(parallel_cfg.get("dp_replicate_size", 1))
    return max(1, tp * pp * dp_shard * dp_rep)


CUR_DIR = os.path.dirname(os.path.abspath(__file__))
PYTHON = sys.executable
CTRL_ENTRY = "cosmos_rl.dispatcher.run_web_panel"
POLICY_ENTRY = "cosmos_rl.policy.train"
ROLLOUT_ENTRY = "cosmos_rl.rollout.rollout_entrance"


def _launch_controller(cfg_file: str, port: int):
    """Start controller web panel on given port."""
    cmd = f"{PYTHON} -m {CTRL_ENTRY} --config {cfg_file} --port {port}"
    env = dict(os.environ, COSMOS_ROLE="Controller")
    return subprocess.Popen(
        cmd,
        shell=True,
        env=env,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        preexec_fn=os.setsid,
    )


def _launch_torchrun(entry: str, world_size: int, cuda_devices: str | None = None):
    """Launch a torchrun worker running module entry with given world size"""
    cmd = [
        "torchrun",
        f"--nproc_per_node={world_size}",
        "--role=rank",
        "--tee=3",
        "--rdzv_backend=c10d",
        "--rdzv_endpoint=localhost:0",
        "-m",
        entry,
    ]
    env = os.environ.copy()
    if cuda_devices is not None:
        env["CUDA_VISIBLE_DEVICES"] = cuda_devices
    return subprocess.Popen(
        cmd,
        env=env,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        preexec_fn=os.setsid,
    )


def _prepare_cfg(src_cfg: str) -> str:
    """Overwrite dataset.name to local tiny set and dump to temp file."""
    with open(src_cfg, "r") as f:
        cfg = toml.load(f)
    tmp = tempfile.NamedTemporaryFile(suffix=".toml", delete=False, mode="w+")
    toml.dump(cfg, tmp)
    tmp.close()
    return tmp.name


def _kill_tree(proc):
    """Kill *proc* and all its children recursively."""
    if proc is None:
        return
    if psutil is not None:
        try:
            parent = psutil.Process(proc.pid)
            for child in parent.children(recursive=True):
                try:
                    child.kill()
                except psutil.NoSuchProcess:
                    pass
            parent.kill()
        except psutil.NoSuchProcess:
            pass
    else:
        try:
            os.killpg(os.getpgid(proc.pid), signal.SIGKILL)
        except ProcessLookupError:
            pass


# Stream flag
STREAM_LOGS = False
if "--stream" in sys.argv or "--verbose" in sys.argv:
    STREAM_LOGS = True
    # remove the flag to avoid unittest/pytest complaining
    sys.argv = [arg for arg in sys.argv if arg not in ("--stream", "--verbose")]


def _reader_thread(proc, bucket):
    """Read lines from proc.stdout, store and optionally echo."""
    for line in iter(proc.stdout.readline, b""):
        text = line.decode("utf-8", errors="ignore")
        bucket.append(text)
        if STREAM_LOGS:
            sys.stdout.write(text)
            sys.stdout.flush()


def run_smoke(cfg_name: str, need_rollout: bool):
    """Launch controller/policy/(rollout) and wait until they finish."""
    cfg_file = _prepare_cfg(os.path.join(CUR_DIR, "configs", cfg_name))

    with open(cfg_file, "r") as f:
        cfg_dict = toml.load(f)

    port = util.find_available_port(13000)

    controller = _launch_controller(cfg_file, port)
    os.environ["COSMOS_CONTROLLER_HOST"] = f"localhost:{port}"

    policy_ws = _calc_world_size(cfg_dict["policy"]["parallelism"], is_rollout=False)
    total_gpus = torch.cuda.device_count()

    if policy_ws > total_gpus:
        raise RuntimeError(
            f"Policy requires {policy_ws} GPUs but only {total_gpus} are visible"
        )

    policy_dev = ",".join(str(i) for i in range(policy_ws))

    rollout_ws = None
    rollout_dev = None
    if need_rollout:
        rollout_ws = _calc_world_size(
            cfg_dict["rollout"]["parallelism"], is_rollout=True
        )
        if policy_ws + rollout_ws > total_gpus:
            raise RuntimeError(
                f"Total GPUs required (policy {policy_ws} + rollout {rollout_ws}) exceeds available {total_gpus}"
            )
        rollout_dev = ",".join(str(i) for i in range(policy_ws, policy_ws + rollout_ws))

    policy = _launch_torchrun(
        POLICY_ENTRY, world_size=policy_ws, cuda_devices=policy_dev
    )
    rollout = (
        _launch_torchrun(ROLLOUT_ENTRY, world_size=rollout_ws, cuda_devices=rollout_dev)
        if need_rollout
        else None
    )

    procs = [p for p in (controller, policy, rollout) if p]

    # Start readers
    controller_logs: list[str] = []
    threading.Thread(
        target=_reader_thread, args=(controller, controller_logs), daemon=True
    ).start()
    policy_logs: list[str] = []
    threading.Thread(
        target=_reader_thread, args=(policy, policy_logs), daemon=True
    ).start()
    rollout_logs: list[str] = []
    if rollout:
        threading.Thread(
            target=_reader_thread, args=(rollout, rollout_logs), daemon=True
        ).start()

    timeout_sec = 1000
    try:
        # Wait for policy / rollout to finish
        for worker in (policy, rollout):
            if worker is None:
                continue
            try:
                worker.wait(timeout=timeout_sec)
            except subprocess.TimeoutExpired:
                _kill_tree(worker)
                raise RuntimeError(f"{worker.args} timed out after {timeout_sec}s")
            if worker.returncode != 0:
                raise RuntimeError(
                    f"{worker.args} exited with code {worker.returncode}"
                )

        # Controller should finish after workers
        if controller.poll() is None:
            controller.terminate()
        controller.wait(timeout=30)

        # Compose logs
        captured_logs = policy_logs + rollout_logs + controller_logs
        all_logs = "\n".join(captured_logs)

        try:
            if need_rollout:
                reward_vals = [
                    float(m)
                    for m in re.findall(
                        r"Reward\s+Max\s*:\s*([0-9eE+\-.]+)",
                        all_logs,
                        flags=re.IGNORECASE,
                    )
                ]
                assert reward_vals, "No Reward Max found in logs"
                assert any(v > 0 for v in reward_vals), "All Reward Max are 0"
                print(
                    f"[INTEGRATION TEST] GRPO reward check passed, Reward Max list: {reward_vals}"
                )
            else:
                loss_vals = [
                    float(m)
                    for m in re.findall(r"Validation loss:\s*([0-9eE+\-.]+)", all_logs)
                ]
                assert loss_vals, "No Validation loss found in logs"
                max_loss = max(loss_vals)
                assert max_loss <= 1.3, f"Validation loss too high: {max_loss} > 1.3"
                print(
                    f"[INTEGRATION TEST] SFT validation check passed, Validation loss: {max_loss}"
                )
        except AssertionError:
            print("\n=== Captured logs (tail) ===")
            print(all_logs[-5000:])
            raise
    finally:
        for p in procs:
            if p.poll() is None:
                _kill_tree(p)
        for p in procs:
            try:
                p.wait(timeout=10)
            except subprocess.TimeoutExpired:
                _kill_tree(p)


if __name__ == "__main__":
    cases = [
        ("sft_integration_test.toml", False),  # SFT
        ("grpo_integration_test.toml", True),  # GRPO
    ]

    total_start = time.time()
    for cfg, need_rollout in cases:
        case_start = time.time()
        print(f"[INTEGRATION TEST] Running {cfg}  (rollout={need_rollout})")
        run_smoke(cfg, need_rollout)
        case_elapsed = time.time() - case_start
        print(f"[INTEGRATION TEST] ✅  {cfg} finished in {case_elapsed:.1f} seconds")
    total_elapsed = time.time() - total_start
    print(f"[INTEGRATION TEST] All cases finished in {total_elapsed:.1f} seconds")



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from torch import nn
from typing import Tuple, List, Optional, Callable
from transformers import AutoConfig, AutoModelForCausalLM
from cosmos_rl.utils.util import (
    sync_model_vocab,
    clear_weight_name,
    retry,
)
from cosmos_rl.utils.constant import COSMOS_HF_MODEL_TYPES
from cosmos_rl.policy.model.base import BaseModel, ModelRegistry
from cosmos_rl.utils.logging import logger
from cosmos_rl.policy.model.hf_llm.weight_converter import convert_weight_from_hf
from cosmos_rl.policy.model.hf_llm.weight_mapper import HFLLMWeightMapper
from cosmos_rl.utils.parallelism import ParallelDims
from cosmos_rl.policy.config import Config as CosmosConfig
from functools import cached_property


@ModelRegistry.register(HFLLMWeightMapper)
class HFLLMModel(BaseModel):
    """
    HFLLM Module

    Args:
        hf_config : Model configuration arguments.
        model: AutoModelForCausalLM model.

    Attributes:
        hf_config : Model configuration arguments.
        model: AutoModelForCausalLM model.
        layers: List of AutoModelForCausalLM blocks.
        src_model_type: Model type.
    """

    @staticmethod
    def supported_model_types():
        return [COSMOS_HF_MODEL_TYPES]

    def __init__(self, hf_config, model):
        super().__init__(hf_config)
        self.hf_config = hf_config
        self.model = model
        self.layers = model.model.layers
        self.src_model_type = hf_config.model_type

    def forward(
        self,
        input_ids: torch.Tensor,
        position_ids: torch.Tensor = None,
        *args,
        **kwargs,
    ):
        out = self.model(
            input_ids,
            position_ids=position_ids,
            is_causal=True,
            attention_mask=None,
            past_key_values=None,
            use_cache=False,
            *args,
            **kwargs,
        )
        return out.logits

    def post_to_empty_hook(self, cosmos_config: CosmosConfig):
        # reset buffer registered in __init__() function,
        # e.g. rotary_emb.inv_freq, embed_tokens.embed_scale
        model = self.model.model
        rotary_emb = getattr(model, "rotary_emb", None)
        if rotary_emb is not None:
            rope_init_fn = getattr(rotary_emb, "rope_init_fn", None)
            if rope_init_fn is not None:
                inv_freq, rotary_emb.attention_scaling = rope_init_fn(
                    self.hf_config, None
                )
                rotary_emb.register_buffer("inv_freq", inv_freq, persistent=False)
            else:
                logger.warning(
                    "rotary_emb does not have rope_init_fn, cannot reset inv_freq."
                )
        # Models like Gemma have rotary_emb_local
        rotary_emb_local = getattr(model, "rotary_emb_local", None)
        if rotary_emb_local is not None:
            rope_init_fn = getattr(rotary_emb_local, "rope_init_fn", None)
            if rope_init_fn is not None:
                local_inv_freq, rotary_emb_local.attention_scaling = rope_init_fn(
                    self.hf_config, None
                )
                rotary_emb_local.register_buffer(
                    "inv_freq", local_inv_freq, persistent=False
                )
            else:
                logger.warning(
                    "rotary_emb_local does not have rope_init_fn, cannot reset inv_freq."
                )

        if self.src_model_type in ["gemma3_text"]:
            embed_tokens = model.embed_tokens
            embed_scale = self.hf_config.hidden_size**0.5
            embed_tokens.register_buffer(
                "embed_scale", torch.tensor(embed_scale), persistent=False
            )

    @property
    def parallelize_fn(self):
        from cosmos_rl.policy.model.hf_llm.parallelize import parallelize

        return parallelize, self

    def apply_pipeline_split(self, pp_rank, pp_size):
        """
        Apply pipeline split to the model.
        This typically involves splitting the model into multiple stages,
        and moving each stage to a different device.
        """
        assert False, "Pipeline split is not supported for HFLLMModel"

    def load_hf_weights(
        self,
        model_name_or_path: str,
        parallel_dims: ParallelDims,
        device: torch.device,
    ):
        """
        Load weights from a HuggingFace model.

        Args:
            model_path (str): Path to the HuggingFace model.
            parallel_dims (ParallelDims): Parallel dimensions definition.
            info_inly (bool): Only collect the tensor infomation without actual data loading.
        """
        model_type = retry(AutoConfig.from_pretrained)(model_name_or_path).model_type
        model_with_weights = AutoModelForCausalLM.from_pretrained(
            model_name_or_path
        ).to(device)
        state_dict = model_with_weights.state_dict()
        self_state_dict = self.model.state_dict()
        self_state_dict = {clear_weight_name(k): v for k, v in self_state_dict.items()}
        all_tensor_names = self_state_dict.keys()
        lm_head_weight_key = "lm_head.weight"
        embed_tokens_weight_key = "model.embed_tokens.weight"
        reserved = {}

        for name, tensor in state_dict.items():
            if name == embed_tokens_weight_key:
                reserved[name] = tensor
            dest_name, shared_weight = convert_weight_from_hf(
                tensor, name, model_type, parallel_dims
            )

            target_tensor = self_state_dict[dest_name]
            is_dist_tensor = isinstance(target_tensor, torch.distributed.tensor.DTensor)
            local_view = target_tensor.to_local() if is_dist_tensor else target_tensor
            assert (
                local_view.shape == shared_weight.shape
            ), f"Shape mismatch: {local_view.shape} != {shared_weight.shape} for {dest_name}"
            with torch.no_grad():
                local_view.data.copy_(shared_weight)

        if (
            lm_head_weight_key not in all_tensor_names
            and embed_tokens_weight_key in all_tensor_names
        ):
            # tied with embed_tokens.weight
            name = lm_head_weight_key
            assert embed_tokens_weight_key in reserved
            tensor = reserved[embed_tokens_weight_key]
            dest_name, shared_weight = convert_weight_from_hf(
                tensor, name, model_type, parallel_dims
            )
            if dest_name in self_state_dict:
                target_tensor = self_state_dict[dest_name]
                is_dist_tensor = isinstance(
                    target_tensor, torch.distributed.tensor.DTensor
                )
                local_view = (
                    target_tensor.to_local() if is_dist_tensor else target_tensor
                )
                assert (
                    local_view.shape == shared_weight.shape
                ), f"Shape mismatch: {local_view.shape} != {shared_weight.shape} for {dest_name}"
                with torch.no_grad():
                    local_view.data.copy_(shared_weight)

    def get_position_ids(self, **kwargs) -> Tuple[torch.Tensor, torch.Tensor, int]:
        seq_dim_idx = 1
        inputs = kwargs["input_ids"]
        position_ids = (
            torch.arange(inputs.size(-1), dtype=torch.long, device=inputs.device)
            .unsqueeze(0)
            .expand_as(inputs)
        )
        return position_ids, inputs, seq_dim_idx

    def separate_model_parts(self) -> List[nn.Module]:
        return [self.model]

    @cached_property
    def _get_nparams_and_flops_fn(self) -> Callable[[int], tuple[int, int]]:
        nparams = sum(p.numel() for p in self.parameters())
        nparams_embedding = sum(
            sum(p.numel() for p in m.parameters())
            for m in self.children()
            if isinstance(m, nn.Embedding)
        )

        # Reasoning behind the factor of 12 for the self-attention part of the formula:
        # 1. each self-attention has 2 matmul in the forward and 4 in the backward (6)
        # 2. the flash attention does 1 more matmul recomputation in the backward
        #    but recomputation should not be counted in calculating MFU           (+0)
        # 3. each matmul performs 1 multiplication and 1 addition                 (*2)
        # 4. we follow the convention and do not account for sparsity in causal attention
        layers, heads, head_dim = (
            len(self.layers),
            self.hf_config.num_attention_heads,
            self.hf_config.hidden_size // self.hf_config.num_attention_heads,
        )
        return lambda seq_len: (
            nparams,
            6 * (nparams - nparams_embedding)
            + 12 * layers * heads * head_dim * seq_len,
        )

    def get_nparams_and_flops(self, seq_len: int) -> tuple[int, int]:
        return self._get_nparams_and_flops_fn(seq_len)

    @classmethod
    def from_model_args(cls, hf_config) -> "HFLLMModel":
        """
        Initialize a HFLLM model from a HFLLMArgs object.

        Args:
            hf_config : hf model config.

        Returns:
            HFLLMModel: HFLLM model.

        """
        model = AutoModelForCausalLM.from_config(hf_config)
        return cls(hf_config, model)

    @classmethod
    def from_pretrained(
        cls,
        hf_config: AutoConfig,
        model_name_or_path: str,
        max_position_embeddings: Optional[int] = None,
    ) -> "HFLLMModel":
        """
        Initialize a HFLLM model from a pretrained model.

        Args:
            hf_config (AutoConfig): HuggingFace config.
            model_name_or_path (str): Model name or path to the pretrained model.
            max_position_embeddings (int): Maximum position embeddings.

        Returns:
            HFLLMModel: HFLLM model.

        """

        if max_position_embeddings is None:
            max_position_embeddings = hf_config.max_position_embeddings
        else:
            hf_config.max_position_embeddings = max_position_embeddings
        _ = sync_model_vocab(model_name_or_path)

        return cls.from_model_args(hf_config)

    def named_parameters(self, *args, **kwargs):
        return self.model.named_parameters(*args, **kwargs)

    @classmethod
    def fqn_filter_for_fp8(cls) -> List[str]:
        return ["lm_head"]

    def check_cp_compatible(self, cp_size: int, tp_size: int):
        if not (self.hf_config.num_attention_heads % (cp_size * tp_size) == 0):
            raise ValueError(
                f"Model is not compatible with cp parallelism, model's head number={self.hf_config.num_attention_heads} is not divisible by cp size({cp_size}) * tp_size({tp_size}) = {cp_size * tp_size}"
            )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import List, Literal
import os
import sys

sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from util import NodeLaunchMetadata, ReplicaLaunchMetadata
import argparse
import math
import tempfile
import subprocess
import json
import toml

logging.basicConfig(level=logging.INFO)


def compute_nodes(
    n_gpu_per_node: int,
    n_gpu_per_replica: int,
    n_replicas: int,
    role: Literal["policy", "rollout"],
) -> List[NodeLaunchMetadata]:
    """
    Compute the number of nodes required for a given number of GPUs per replica and the number of replicas.

    If multiple replicas are colocated on the same node, the visible GPUs for each replica are computed.

    Returns:
        A list of NodeLaunchMetadata, one for each node.
    """
    n_nodes = 0
    rendezvous_port = 29345

    node_launch_metadata = []
    if n_gpu_per_replica >= n_gpu_per_node:
        assert (
            n_gpu_per_replica % n_gpu_per_node == 0
        ), f"Number of GPUs per policy must be a multiple of {n_gpu_per_node}"
        n_policy_nodes = n_replicas * (n_gpu_per_replica // n_gpu_per_node)

        rendezvous_node = 0
        for i_node in range(n_policy_nodes):
            if i_node % (n_gpu_per_replica // n_gpu_per_node) == 0:
                rendezvous_node = i_node

            replica_launch_meta = [
                # Only one replica per node, no colocation or rendezvous conflicts
                ReplicaLaunchMetadata(
                    nnode=n_gpu_per_replica // n_gpu_per_node,
                    role=role,
                    rendezvous_node=rendezvous_node,
                    rendezvous_port=rendezvous_port,
                    visible_gpus=list(range(0, n_gpu_per_node)),
                )
            ]
            node_launch_metadata.append(
                NodeLaunchMetadata(colocation=replica_launch_meta)
            )
    else:
        possible_gpu_per_replica = []
        for divisor in range(1, n_gpu_per_node):
            if n_gpu_per_node % divisor == 0:
                possible_gpu_per_replica.append(divisor)

        assert (
            n_gpu_per_replica in possible_gpu_per_replica
        ), f"Number of GPUs per policy must be one of {possible_gpu_per_replica}, got {n_gpu_per_replica}."
        n_policy_nodes = math.ceil(n_replicas * n_gpu_per_replica / n_gpu_per_node)

        replica_counter = 0
        for i_node in range(n_policy_nodes):
            replica_launch_meta = []
            local_replica_counter = 0
            while replica_counter < n_replicas:
                replica_launch_meta.append(
                    ReplicaLaunchMetadata(
                        nnode=1,
                        role=role,
                        rendezvous_node=i_node,  # Always on the same node
                        rendezvous_port=rendezvous_port
                        + replica_counter,  # To avoid conflicts with other replicas on the same node
                        visible_gpus=list(
                            range(
                                local_replica_counter * n_gpu_per_replica,
                                (local_replica_counter + 1) * n_gpu_per_replica,
                            )
                        ),
                    )
                )
                replica_counter += 1
                local_replica_counter += 1
                if replica_counter == n_replicas:
                    break
                elif local_replica_counter * n_gpu_per_replica >= n_gpu_per_node:
                    # Dispatch left to next node
                    break
            node_launch_metadata.append(
                NodeLaunchMetadata(colocation=replica_launch_meta)
            )
    n_nodes += n_policy_nodes

    return node_launch_metadata


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--job-name", type=str, default="cosmos_job")
    parser.add_argument(
        "--ngpu-per-node", type=int, default=8, help="Number of GPUs per compute node."
    )
    parser.add_argument(
        "--n-policy-replicas",
        type=int,
        default=1,
        help="Number of policy replicas to launch",
    )
    parser.add_argument(
        "--n-rollout-replicas",
        type=int,
        default=1,
        help="Number of rollout replicas to launch",
    )
    parser.add_argument(
        "--slurm-partition", type=str, default="batch", help="SLURM partition to use"
    )
    parser.add_argument(
        "--slurm-account", type=str, default="sw_aidot", help="SLURM account to use"
    )
    parser.add_argument(
        "--config-path",
        type=str,
        required=True,
        help="Path to the controller config file",
    )
    parser.add_argument(
        "--repo-root-path", type=str, default=None, help="Path to the repository root"
    )
    parser.add_argument(
        "--output-root-path", type=str, required=True, help="Path to the output root"
    )
    parser.add_argument(
        "--cosmos-container",
        type=str,
        required=True,
        help="Path to the cosmos container",
    )
    parser.add_argument(
        "--extra-sbatch-args",
        type=str,
        nargs="*",
        default=["--gres=gpu:8"],
        help="Extra #SBATCH arguments",
    )
    parser.add_argument(
        "launcher",
        nargs="?",  # “?” means 0 or 1 occurrences
        default="cosmos_rl.dispatcher.run_web_panel",
        help="The launcher to use, default is `cosmos_rl.dispatcher.run_web_panel`, a custom launcher can be provided for custom dataset and reward functions injection.",
    )

    args = parser.parse_args()

    with open(args.config_path, "r") as f:
        config = toml.load(f)
    min_n_gpus_policy = (
        config["policy"]["parallelism"]["tp_size"]
        * config["policy"]["parallelism"]["pp_size"]
        * config["policy"]["parallelism"]["cp_size"]
    )
    train_type = config["train"]["train_policy"]["type"]

    if "rollout" in config:
        # sft case may not have rollout config
        min_n_gpus_rollout = (
            config["rollout"]["parallelism"]["tp_size"]
            * config["rollout"]["parallelism"]["pp_size"]
        )
    if config["policy"]["parallelism"]["dp_shard_size"] >= 1:
        min_n_gpus_policy = (
            min_n_gpus_policy * config["policy"]["parallelism"]["dp_shard_size"]
        )
    # Update the n_init_replicas in the config
    if "policy" in config and "parallelism" in config["policy"]:
        config["policy"]["parallelism"]["n_init_replicas"] = args.n_policy_replicas
    if "rollout" in config and "parallelism" in config["rollout"]:
        # Only available for RL.
        config["rollout"]["parallelism"]["n_init_replicas"] = args.n_rollout_replicas
    # Create a temporary file and write to it
    cache_dir = os.path.join(os.path.expanduser("~"), ".cache/cosmos_rl/tmp_config")
    os.makedirs(cache_dir, exist_ok=True)
    with tempfile.NamedTemporaryFile(
        dir=cache_dir, mode="w+", suffix=".toml", delete=False
    ) as tmpfile:
        toml.dump(config, tmpfile)
        config_tmpfile = tmpfile.name
        logging.info(f"Config written to {config_tmpfile}")

    policy_node_launch_metadata: List[NodeLaunchMetadata] = compute_nodes(
        args.ngpu_per_node, min_n_gpus_policy, args.n_policy_replicas, "policy"
    )
    n_policy_nodes = len(policy_node_launch_metadata)

    if train_type == "sft":
        rollout_node_launch_metadata = []
    else:
        rollout_node_launch_metadata: List[NodeLaunchMetadata] = compute_nodes(
            args.ngpu_per_node, min_n_gpus_rollout, args.n_rollout_replicas, "rollout"
        )
    n_rollout_nodes = len(rollout_node_launch_metadata)

    # Template for the slurm script
    template_vars = {
        "TOTAL_NODES": f"{n_policy_nodes + n_rollout_nodes}",
        "OUTPUT_ROOT_PATH": args.output_root_path,
        "COSMOS_CONTAINER": args.cosmos_container,
        "SLURM_PARTITION": args.slurm_partition,
        "SLURM_ACCOUNT": args.slurm_account,
        "SLURM_JOB_NAME": args.job_name,
        "CONFIG_PATH": config_tmpfile,
        "LAUNCHER": args.launcher,
        "EXTRA_SBATCH_ARGS": "\n".join(
            f"#SBATCH {arg}" for arg in args.extra_sbatch_args
        ),
    }

    # Environment variables
    env_vars = {
        "NUM_POLICY_NODES": f"{n_policy_nodes}",
        "NUM_ROLLOUT_NODES": f"{n_rollout_nodes}",
        "TOTAL_NODES": f"{n_policy_nodes + n_rollout_nodes}",
    }

    if args.repo_root_path is not None:
        env_vars["REPO_ROOT_PATH"] = args.repo_root_path

    # Read the template relative to the current file
    with open(
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "cosmos_rl_job_multi_node.sh"
        ),
        "r",
    ) as f:
        template = f.read()

    # Replace the template variables
    for key, value in template_vars.items():
        template = template.replace(f"[[{key}]]", value)

    # Write the template to temp file
    temp_file = tempfile.NamedTemporaryFile(delete=False)
    temp_file.write(template.encode("utf-8"))
    temp_file.close()
    logging.info(f"Template written to {temp_file.name}")

    env = os.environ.copy()
    env.update(env_vars)
    env["NODE_LAUNCH_METADATA_POLICY"] = json.dumps(
        [x.to_json() for x in policy_node_launch_metadata]
    )
    env["NODE_LAUNCH_METADATA_ROLLOUT"] = json.dumps(
        [x.to_json() for x in rollout_node_launch_metadata]
    )
    proc = subprocess.Popen(["sbatch", temp_file.name], env=env)
    proc.wait()
    if proc.returncode != 0:
        logging.error(f"Failed to submit job: {proc.returncode}")
        sys.exit(1)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Optional, Any, List, Dict
from torch.utils.data import Dataset, ConcatDataset
from datasets import load_dataset
from cosmos_rl.launcher.worker_entry import main as launch_worker
import cosmos_rl.utils.util as util
from cosmos_rl.policy.config import Config
from cosmos_rl.dispatcher.algo.reward import single_choice_reward_fn, format_reward_fn
from transformers import AutoTokenizer
from cosmos_rl.utils.util import basename_from_modelpath
from cosmos_rl.dispatcher.data.packer import DataPacker, Qwen2_5_VLM_DataPacker
from cosmos_rl.utils.logging import logger
from cosmos_rl.policy.config import Config as CosmosConfig
import argparse
import toml

FPS = 1
MAX_PIXELS = 81920


class CosmosGRPODataset(Dataset):
    def get_mm_files_paths(self, dataset_name: str, dataset_subset: str):
        cosmos_cache_dir = os.environ.get(
            "COSMOS_CACHE", os.path.join(os.path.expanduser("~"), ".cache/cosmos/")
        )
        video_clips_path = os.path.join(
            cosmos_cache_dir,
            "datasets",
            basename_from_modelpath(dataset_name),
            dataset_subset,
            "video_clips",
        )
        if not os.path.exists(video_clips_path):
            raise FileNotFoundError(
                f"Dataset directory {video_clips_path} does not exist. Please check the dataset path."
            )
        mm_files_paths = {}
        for root, _, files in os.walk(video_clips_path):
            for file in files:
                if file.endswith((".mp4", ".avi", ".mov")):  # Common video extensions
                    mm_files_paths[file] = os.path.join(root, file)
        return mm_files_paths

    def setup(self, config: CosmosConfig, tokenizer: AutoTokenizer, *args, **kwargs):
        self.config = config
        self.tokenizer = tokenizer
        self.dataset = load_dataset(
            config.train.train_policy.dataset.name,
            config.train.train_policy.dataset.subset,
        )
        if config.train.train_policy.dataset.split:
            if isinstance(config.train.train_policy.dataset.split, list):
                dataset_list = []
                for split_name in config.train.train_policy.dataset.split:
                    dataset_list.append(self.dataset[split_name])
                self.dataset = ConcatDataset(dataset_list)
            else:
                assert isinstance(config.train.train_policy.dataset.split, str)
                self.dataset = self.dataset[config.train.train_policy.dataset.split]
        self.mm_files_paths = self.get_mm_files_paths(
            config.train.train_policy.dataset.name,
            config.train.train_policy.dataset.subset,
        )

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx: int) -> tuple[str, str]:
        """
        Return a tuple of (prompt, reference answer)
        """
        payload = self.dataset[idx]

        system_prompt = ""

        choices = payload["qa_pairs"]["index2ans"]
        user_prompt = (
            payload["qa_pairs"]["question"]
            + "\n"
            + "\n".join([f"({i}) {choice}" for i, choice in choices.items()])
        )
        user_prompt += (
            "\nAnswer with the option's letter from the given choices directly."
        )
        user_prompt += "\nPlease answer the question in the following format: <think> your reasoning </think> <answer> your answer </answer>."

        user_conv = [
            {
                "type": "text",
                "text": user_prompt,
            },
        ]

        if "video" in payload or "image" in payload:
            if "video" in payload:
                multi_modal_content = {
                    "type": "video",
                    "video": self.mm_files_paths[payload["video"].split("/")[-1]],
                    "max_pixels": MAX_PIXELS,
                    "fps": FPS,
                }
            else:
                multi_modal_content = {
                    "type": "image",
                    "image": self.mm_files_paths[payload["image"].split("/")[-1]],
                }
            user_conv.insert(0, multi_modal_content)

        conversations = [
            {
                "role": "system",
                "content": system_prompt,
            },
            {
                "role": "user",
                "content": user_conv,
            },
        ]

        return conversations

    def get_reference_answer(self, idx: int) -> str:
        return self.dataset[idx]["qa_pairs"]["answer"]


class CosmosGRPOValDataset(CosmosGRPODataset):
    """
    This is a validation dataset for Cosmos GRPO, which is used to evaluate the performance of the model.
    It should be used in the launcher to evaluate the model during training.
    """

    def setup(self, config: Config, tokenizer: AutoTokenizer, *args, **kwargs):
        if not config.train.enable_validation:
            logger.warning(
                "Validation is not enabled in the config. Skipping setup for CosmosGRPOValDataset."
            )
            return

        self.config = config
        self.tokenizer = tokenizer
        self.dataset = load_dataset(
            config.validation.dataset.name, config.validation.dataset.subset
        )

        if config.validation.dataset.split:
            if isinstance(config.validation.dataset.split, list):
                dataset_list = []
                for split_name in config.validation.dataset.split:
                    dataset_list.append(self.dataset[split_name])
                self.dataset = ConcatDataset(dataset_list)
            else:
                assert isinstance(config.validation.dataset.split, str)
                self.dataset = self.dataset[config.validation.dataset.split]
        # Prepare the data for Cosmos GRPO
        # This is a hack to make the dataset compatible with the training data
        # Change the training dataset name and subset to utilize the same data preparation logic
        self.mm_files_paths = self.get_mm_files_paths(
            config.validation.dataset.name,
            config.validation.dataset.subset,
        )


def custom_reward_fn(
    to_be_evaluated: str, reference: Optional[str] = None, *args, **kwargs
) -> float:
    return sum(
        [
            single_choice_reward_fn(to_be_evaluated, reference, *args, **kwargs),
            format_reward_fn(to_be_evaluated, reference, *args, **kwargs),
        ]
    )


def custom_val_reward_fn(
    to_be_evaluated: str, reference: Optional[str] = None, *args, **kwargs
) -> float:
    return single_choice_reward_fn(to_be_evaluated, reference, *args, **kwargs)


class DemoDataPacker(DataPacker):
    """
    This is a demo data packer that wraps the underlying data packer of the selected model.
    This is meaningless for this example, but useful for explaining:
        - how dataset data is processed and collated into a mini-batch for rollout engine;
        - how rollout output is processed and collated into a mini-batch for policy model;
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Check source code of Qwen2_5_VLM_DataPacker to see how it's implemented
        self.underlying_data_packer = Qwen2_5_VLM_DataPacker()

    def setup(self, config: Config, tokenizer: AutoTokenizer, *args, **kwargs):
        """
        This method is optional and get called by launcher after being mounted
        `config`: config;
        `tokenizer`: tokenizer;
        """
        super().setup(config, tokenizer, *args, **kwargs)
        self.underlying_data_packer.setup(config, tokenizer, *args, **kwargs)

    def get_rollout_input(self, item: Any) -> Any:
        """
        Convert dataset item into what rollout engine (e.g. vllm) expects
        """
        return self.underlying_data_packer.get_rollout_input(item)

    def rollout_collate_fn(self, items: List[Any]) -> Any:
        """
        Collate the rollout inputs into a mini-batch for rollout engine
        """
        return self.underlying_data_packer.rollout_collate_fn(items)

    def get_policy_input(
        self, item: Any, rollout_output: str, n_ignore_prefix_tokens: int = 0
    ) -> Any:
        """
        Process samples & rollout output before collating them into a mini-batch
        """
        return self.underlying_data_packer.get_policy_input(
            item, rollout_output, n_ignore_prefix_tokens
        )

    def policy_compute_max_len(self, processed_samples: List[Any]) -> int:
        """
        Compute the maximum sequence length of the mini-batch
        """
        return self.underlying_data_packer.policy_compute_max_len(processed_samples)

    def policy_collate_fn(
        self, processed_samples: List[Any], computed_max_len: int
    ) -> Dict[str, Any]:
        """
        Collate the mini-batch into the kwargs required by the policy model
        """
        return self.underlying_data_packer.policy_collate_fn(
            processed_samples, computed_max_len
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--config", type=str, required=True)
    args = parser.parse_known_args()[0]
    with open(args.config, "r") as f:
        config = toml.load(f)
    config = Config.from_dict(config)

    util.prepare_cosmos_data(
        dataset=config.train.train_policy.dataset, fps=FPS, max_pixels=MAX_PIXELS
    )
    if config.train.enable_validation:
        util.prepare_cosmos_data(
            dataset=config.validation.dataset, fps=FPS, max_pixels=MAX_PIXELS
        )

    # It is best practice to pass the dataset and val_dataset as factory functions
    # so that the dataset and val_dataset can be loaded on demand. (Not all workers need them)
    def get_dataset(config: CosmosConfig) -> Dataset:
        return CosmosGRPODataset()

    def get_val_dataset(config: CosmosConfig) -> Dataset:
        return CosmosGRPOValDataset() if config.train.enable_validation else None

    launch_worker(
        dataset=get_dataset,
        reward_fns=[custom_reward_fn],
        data_packer=DemoDataPacker(),
        val_dataset=get_val_dataset,
        val_reward_fns=[custom_val_reward_fn],
        val_data_packer=DemoDataPacker(),
    )



from cosmos_rl.dispatcher.data.packer.base import DataPacker
from typing import List, Any, Dict, Optional
import torch
import torch.nn.functional as F
from cosmos_rl.utils.util import retry
from cosmos_rl.policy.config import Config
from transformers import AutoTokenizer, AutoProcessor, AutoConfig
from qwen_vl_utils import process_vision_info
import logging

IGNORE_LABEL_ID = -100


class Qwen2_5_VLM_DataPacker(DataPacker):
    """
    Data protocol & processing logic for the Qwen2.5 VLM for SFT and RL training.
    """

    Payload = List[Dict[str, Any]]

    class RLPolicyInput:
        input_ids: List[int]
        logprob_masks: List[int]

        def __init__(self, input_ids: List[int], logprob_masks: List[int]):
            self.input_ids = input_ids
            self.logprob_masks = logprob_masks

    def setup(self, config: Config, tokenizer: AutoTokenizer, *args, **kwargs):
        super().setup(config, tokenizer, *args, **kwargs)
        self.hf_processor = retry(AutoProcessor.from_pretrained)(
            config.policy.model_name_or_path
        )
        # disable qwen_vl_utils logger
        qwen_vl_utils_logger = logging.getLogger("qwen_vl_utils")
        qwen_vl_utils_logger.setLevel(logging.WARNING)
        qwen_vl_utils_logger.propagate = False

        hf_config = retry(AutoConfig.from_pretrained)(config.policy.model_name_or_path)
        if hasattr(hf_config, "image_token_id"):
            self.image_token_id = hf_config.image_token_id
        else:
            self.image_token = None
            self.image_token_id = None
        if hasattr(hf_config, "video_token_id"):
            self.video_token_id = hf_config.video_token_id
        else:
            self.video_token = None
            self.video_token_id = None
        self.vision_ids = [self.image_token_id, self.video_token_id]
        self.hf_config = hf_config

    def get_rollout_input(self, sample: Payload) -> Any:
        """
        This VL data packer only accepts the conversation data format.
        check https://github.com/QwenLM/Qwen2.5-VL?tab=readme-ov-file#using---transformers-to-chat for more details.

        example:
        [
            {
                "role": "user",
                "content": [
                    {
                        "type": "image",
                        "image": "https://qianwen-res.oss-cn-beijing.aliyuncs.com/Qwen-VL/assets/demo.jpeg",
                    },
                    {"type": "text", "text": "Describe this image."},
                ],
            }
        ]

        It is user's responsibility to ensure the conversation format is correct
          and multi-media files involved in conversation are accessible.
        """
        assert all(
            isinstance(x, dict) and "role" in x and "content" in x for x in sample
        ), "All samples should be in conversation format, but got: {}".format(sample)

        # Here we need to convert the conversation format to the format required by vllm
        prompt = self.hf_processor.apply_chat_template(
            sample, tokenize=False, add_generation_prompt=True
        )
        image_inputs, video_inputs = process_vision_info(sample)

        if video_inputs:
            return {
                "prompt": prompt,
                "multi_modal_data": {"video": video_inputs},
            }
        elif image_inputs:
            return {
                "prompt": prompt,
                "multi_modal_data": {"image": image_inputs},
            }
        else:
            return {
                "prompt": prompt,
            }

    def _replace_assistant_content(
        self,
        token_ids: List[int],
        label_ids: List[int],
        pad_token_id: int,
        eos_token_id: int,
        replacement_ids: List[int],
        pad_run_length: int = 10,
    ) -> List[int]:
        """
        Find the first run of exactly `pad_run_length` pad_token_id's in token_ids,
        replace that run with replacement_ids, and return the new list.
        If no such run is found, returns the original list unchanged.
        """
        n = len(token_ids)
        target_run = [pad_token_id] * pad_run_length

        # find the start index of the first matching run
        for i in range(n - pad_run_length + 1):
            if token_ids[i : i + pad_run_length] == target_run:
                # splice in the replacement
                if (
                    len(token_ids) > i + pad_run_length
                    and token_ids[i + pad_run_length] == eos_token_id
                ):
                    label_ids = (
                        label_ids[:i]
                        + replacement_ids
                        + [eos_token_id]
                        + label_ids[i + pad_run_length + 1 :]
                    )
                else:
                    label_ids = (
                        label_ids[:i]
                        + replacement_ids
                        + label_ids[i + pad_run_length :]
                    )
                return (
                    True,
                    token_ids[:i] + replacement_ids + token_ids[i + pad_run_length :],
                    label_ids,
                )
        # no match found
        return False, token_ids, label_ids

    def _process_single_sample(
        self, conversation: "Qwen2_5_VLM_DataPacker.Payload"
    ) -> Dict[str, Any]:
        try:
            # Replace all the assistant content with consecutive `pad_token` * 10
            pad_token = self.tokenizer.pad_token
            pad_token_id = self.tokenizer.pad_token_id
            eos_token_id = self.tokenizer.eos_token_id
            pad_run_length = 10
            assistant_content = []
            for message in conversation:
                if message["role"] == "assistant":
                    content = message["content"]
                    if isinstance(content, str):
                        assistant_content.append(content)
                    elif isinstance(content, dict):
                        assert "text" in content, f"text not in content: {content}"
                        assistant_content.append(content["text"])
                    else:
                        raise ValueError(f"Unsupported content type: {type(content)}")
                    message["content"] = pad_token * pad_run_length

            prompt = self.hf_processor.apply_chat_template(
                conversation,
                tokenize=False,
                add_generation_prompt=False,
            )
            image_inputs, video_inputs = process_vision_info(conversation)
            inputs = self.hf_processor(
                text=[prompt],
                images=image_inputs,
                videos=video_inputs,
                padding=True,
                return_tensors="pt",
            )

            input_ids = inputs["input_ids"][0].tolist()
            label_ids = [IGNORE_LABEL_ID] * len(input_ids)

            for assistant_content in assistant_content:
                replacement_ids = self.tokenizer.encode(
                    assistant_content, add_special_tokens=False
                )
                replaced, input_ids, label_ids = self._replace_assistant_content(
                    input_ids,
                    label_ids,
                    pad_token_id=pad_token_id,
                    eos_token_id=eos_token_id,
                    replacement_ids=replacement_ids,
                    pad_run_length=pad_run_length,
                )
                if not replaced:
                    raise ValueError("No assistant content to replace")
                if len(input_ids) != len(label_ids):
                    raise ValueError(
                        f"input_ids and label_ids should have the same length, but got {len(input_ids)} and {len(label_ids)}"
                    )
        except Exception as e:
            print(f"Error processing sample: {e}, please fix to ensure SFT works")
            raise e

        result_dict = {
            "input_ids": input_ids,
            "label_ids": label_ids,
        }
        if "pixel_values_videos" in inputs:
            result_dict["pixel_values_videos"] = inputs["pixel_values_videos"]
            result_dict["video_grid_thw"] = inputs["video_grid_thw"]
            result_dict["second_per_grid_ts"] = torch.tensor(
                inputs["second_per_grid_ts"], dtype=torch.float
            )
            result_dict["pixel_values_videos_lengths_per_sample"] = inputs[
                "pixel_values_videos"
            ].shape[0]

        if "pixel_values" in inputs:
            result_dict["pixel_values"] = inputs["pixel_values"]
            result_dict["image_grid_thw"] = inputs["image_grid_thw"]
            result_dict["pixel_values_lengths_per_sample"] = inputs[
                "pixel_values"
            ].shape[0]

        return result_dict

    def _collate_fn(
        self, processed_samples: List[Dict[str, Any]], computed_max_len: int
    ) -> Dict[str, Any]:
        pixel_values_videos = [x["pixel_values_videos"] for x in processed_samples]
        video_grid_thw = [x["video_grid_thw"] for x in processed_samples]
        second_per_grid_ts = [x["second_per_grid_ts"] for x in processed_samples]
        pixel_values = [x["pixel_values"] for x in processed_samples]
        image_grid_thw = [x["image_grid_thw"] for x in processed_samples]
        pixel_values_videos_lengths_per_sample = [
            x["pixel_values_videos_lengths_per_sample"] for x in processed_samples
        ]
        pixel_values_lengths_per_sample = [
            x["pixel_values_lengths_per_sample"] for x in processed_samples
        ]

        if all([x is not None for x in pixel_values_videos]):
            assert all(
                [x is not None for x in pixel_values_videos]
            ), "pixel_values_videos should not be None"
            max_len = max([x.shape[0] for x in pixel_values_videos])
            for i in range(len(pixel_values_videos)):
                pixel_values_videos[i] = pixel_values_videos[i].unsqueeze(0)
                assert (
                    pixel_values_videos[i].ndim == 3
                ), f"pixel_values_videos[i].ndim: {pixel_values_videos[i].ndim}"
                pixel_values_videos[i] = F.pad(
                    pixel_values_videos[i],
                    (0, 0, 0, max_len - pixel_values_videos[i].shape[1]),
                )
            pixel_values_videos = torch.cat(pixel_values_videos, dim=0)
            video_grid_thw = torch.cat(video_grid_thw, dim=0)
            second_per_grid_ts = torch.cat(second_per_grid_ts, dim=0)
        else:
            # TODO(jiaxin): handle the case when there is mixed input: some with video, some without video
            assert all(
                [x is None for x in pixel_values_videos]
            ), "pixel_values_videos should be None"
            pixel_values_videos = None
            video_grid_thw = None
            second_per_grid_ts = None
            pixel_values_videos_lengths_per_sample = None

        if all([x is not None for x in pixel_values]):
            max_len = max([x.shape[0] for x in pixel_values])
            for i in range(len(pixel_values)):
                pixel_values[i] = pixel_values[i].unsqueeze(0)
                assert (
                    pixel_values[i].ndim == 3
                ), f"pixel_values[i].ndim: {pixel_values[i].ndim}"
                pixel_values[i] = F.pad(
                    pixel_values[i],
                    (0, 0, 0, max_len - pixel_values[i].shape[1]),
                )
            pixel_values = torch.cat(pixel_values, dim=0)
            image_grid_thw = torch.cat(image_grid_thw, dim=0)
        else:
            # TODO(jiaxin): handle the case when there is mixed input: some with image, some without image
            assert all([x is None for x in pixel_values]), "pixel_values should be None"
            pixel_values = None
            image_grid_thw = None
            pixel_values_lengths_per_sample = None

        # Shape description:
        #
        # pixel_values_[videos/images]: (BATCH_SIZE, N_PATCH, HIDDEN_SIZE)
        # [video/image]_grid_thw: (BATCH_SIZE, 3)
        # second_per_grid_ts: (BATCH_SIZE, 1)
        # pixel_values_[videos/images]_lengths_per_sample: (BATCH_SIZE, 1)
        batch = {}
        if pixel_values_videos is not None:
            batch["pixel_values_videos"] = pixel_values_videos
            batch["video_grid_thw"] = video_grid_thw
            batch["second_per_grid_ts"] = second_per_grid_ts
            batch["pixel_values_videos_lengths_per_sample"] = torch.tensor(
                pixel_values_videos_lengths_per_sample, dtype=torch.long
            ).view(-1, 1)

        if pixel_values is not None:
            batch["pixel_values"] = pixel_values
            batch["image_grid_thw"] = image_grid_thw
            batch["pixel_values_lengths_per_sample"] = torch.tensor(
                pixel_values_lengths_per_sample, dtype=torch.long
            ).view(-1, 1)

        # Pad the input_ids, logprob_masks
        batch["input_ids"] = torch.tensor(
            [
                x["input_ids"][:computed_max_len]
                + [self.tokenizer.pad_token_id]
                * (max(0, computed_max_len - len(x["input_ids"])))
                for x in processed_samples
            ],
            dtype=torch.long,
        )
        if "label_ids" in processed_samples[0]:
            batch["label_ids"] = torch.tensor(
                [
                    x["label_ids"][:computed_max_len]
                    + [IGNORE_LABEL_ID]
                    * (max(0, computed_max_len - len(x["label_ids"])))
                    for x in processed_samples
                ],
                dtype=torch.long,
            )
        batch["logprob_masks"] = torch.tensor(
            [
                x["logprob_masks"][:computed_max_len]
                + [0] * (max(0, computed_max_len - len(x["logprob_masks"])))
                for x in processed_samples
            ],
            dtype=torch.bool,
        )

        assert len(batch["input_ids"]) == len(
            batch["logprob_masks"]
        ), "The length of input_ids, logprob_masks should be the same"

        return batch

    def get_policy_input(
        self,
        sample: "Qwen2_5_VLM_DataPacker.Payload",
        rollout_output: Optional[str] = None,
        n_ignore_prefix_tokens: int = 0,
    ) -> Any:
        assert all(
            isinstance(x, dict) and "role" in x and "content" in x for x in sample
        ), "All samples should be in conversation format, but got: {}".format(sample)

        x = self._process_single_sample(sample)

        return_dict = {}
        if "pixel_values_videos" in x:
            return_dict["pixel_values_videos"] = x["pixel_values_videos"]
            return_dict["video_grid_thw"] = x["video_grid_thw"]
            return_dict["second_per_grid_ts"] = x["second_per_grid_ts"]
            return_dict["pixel_values_videos_lengths_per_sample"] = x[
                "pixel_values_videos_lengths_per_sample"
            ]
        else:
            return_dict["pixel_values_videos"] = None
            return_dict["video_grid_thw"] = None
            return_dict["second_per_grid_ts"] = None
            return_dict["pixel_values_videos_lengths_per_sample"] = None

        if "pixel_values" in x:
            return_dict["pixel_values"] = x["pixel_values"]
            return_dict["image_grid_thw"] = x["image_grid_thw"]
            return_dict["pixel_values_lengths_per_sample"] = x[
                "pixel_values_lengths_per_sample"
            ]
        else:
            return_dict["pixel_values"] = None
            return_dict["image_grid_thw"] = None
            return_dict["pixel_values_lengths_per_sample"] = None

        # Common fields
        input_ids = x["input_ids"]
        completion_ids = []
        if rollout_output:
            completion_ids = self.tokenizer(rollout_output).input_ids  # Don't pad yet

        return_dict["input_ids"] = input_ids + completion_ids

        return_dict["logprob_masks"] = (
            [0] * (len(input_ids) - 1 + n_ignore_prefix_tokens)
            + [1] * (len(completion_ids) - n_ignore_prefix_tokens)
            + [0]
        )

        # TODO(jiaxin): this is special for SFT, will be removed in ``policy_collate_fn``
        return_dict["label_ids"] = x["label_ids"]
        return return_dict

    def policy_compute_max_len(self, processed_samples: List[Dict[str, Any]]) -> int:
        return max([len(x["input_ids"]) for x in processed_samples])

    def policy_collate_fn(
        self, processed_samples: List[Dict[str, Any]], computed_max_len: int
    ) -> Dict[str, Any]:
        for x in processed_samples:
            if "label_ids" in x:
                del x["label_ids"]
        return self._collate_fn(processed_samples, computed_max_len)

    def sft_process_sample(
        self, sample: "Qwen2_5_VLM_DataPacker.Payload"
    ) -> Dict[str, Any]:
        """
        Accepts either raw text or conversation format.
        """
        return self.get_policy_input(sample)

    def sft_compute_max_len(self, processed_samples: List[Dict[str, Any]]) -> int:
        """
        Compute the maximum sequence length of the processed samples
        """
        return max([len(x["input_ids"]) for x in processed_samples])

    def sft_collate_fn(
        self,
        processed_samples: List[Dict[str, Any]],
        computed_max_len: int,
        pad_token_id: int,
        ignore_label_id: int,
    ) -> Dict[str, Any]:
        # Reuse the RL collate minibatch function
        model_inputs: Dict[str, Any] = self._collate_fn(
            processed_samples, computed_max_len
        )
        del model_inputs["logprob_masks"]
        # Mask the loss on vision padding tokens
        if self.vision_ids is not None:
            assert isinstance(self.vision_ids, list)
            for vision_id in self.vision_ids:
                if vision_id is not None:
                    model_inputs["label_ids"][
                        model_inputs["label_ids"] == vision_id
                    ] = ignore_label_id

        return model_inputs



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import argparse
import uvicorn
import toml
from fastapi import FastAPI
from contextlib import asynccontextmanager
from torch.utils.data import Dataset
import asyncio
import base64
import cloudpickle


from fastapi.responses import HTMLResponse, JSONResponse
from typing import Dict, List, Optional, Callable, Tuple, Union
from cosmos_rl.dispatcher.controller import Controller
import cosmos_rl.utils.constant as constant
from cosmos_rl.dispatcher.protocol import MESH_NAMES
from cosmos_rl.dispatcher.replica import Atom, RolloutGroup, Rollout, Replica
from cosmos_rl.dispatcher.protocol import (
    RegisterRequest,
    ErrorResponse,
    RolloutRequest,
    ValidationReportRequest,
    HandshakeInitiatorRequest,
    HandshakeAcceptorRequest,
    UnregisterRequest,
    TrainAckRequest,
    HeartbeatRequest,
    SetProfileRequest,
    SetTracePathRequest,
    NcclErrRequest,
    NcclStoreClearRequest,
    GetShardSendRecvInstsRequest,
)
from cosmos_rl.policy.config import Config as CosmosConfig
import cosmos_rl.utils.util as util
from cosmos_rl.utils.logging import logger
from cosmos_rl.utils.constant import COSMOS_ROLLOUT_SCAN_INTERVAL
from cosmos_rl.utils.api_suffix import (
    COSMOS_API_PANEL_SUFFIX,
    COSMOS_API_STATUS_SUFFIX,
    COSMOS_API_META_SUFFIX,
    COSMOS_API_REGISTER_SUFFIX,
    COSMOS_API_SET_PROFILE_SUFFIX,
    COSMOS_API_SET_TRACE_PATH_SUFFIX,
    COSMOS_API_UNREGISTER_SUFFIX,
    COSMOS_API_HEARTBEAT_SUFFIX,
    COSMOS_API_NCCL_COMM_INITIATOR_SUFFIX,
    COSMOS_API_NCCL_COMM_ACCEPTOR_SUFFIX,
    COSMOS_API_NCCL_COMM_GET_ALL_SUFFIX,
    COSMOS_API_NCCL_COMM_ERROR_SUFFIX,
    COSMOS_API_NCCL_COMM_STORE_CLEAR_SUFFIX,
    COSMOS_API_NEXT_PROMPT_SUFFIX,
    COSMOS_API_ROLLOUT_SUFFIX,
    COSMOS_API_VALIDATION_REPORT_SUFFIX,
    COSMOS_API_POLICY_TRAIN_ACK_SUFFIX,
    COSMOS_API_POLICY_SHARD_INFOS_SUFFIX,
    COSMOS_API_ROLLOUT_SHARD_INFOS_SUFFIX,
    COSMOS_API_POLICY_SHARD_SEND_INSTS_SUFFIX,
    COSMOS_API_ROLLOUT_SHARD_RECV_INSTS_SUFFIX,
)
from cosmos_rl.dispatcher.data.packer.base import DataPacker
from fastapi.responses import Response
from fastapi import Request


def create_error_response(
    code: int, message: str, status_code: Optional[int] = None
) -> JSONResponse:
    if status_code is None:
        status_code = code // 100
    return JSONResponse(
        ErrorResponse(message=message, code=code).model_dump(), status_code=status_code
    )


controller = Controller()
server = None


@asynccontextmanager
async def lifespan(app: FastAPI):
    async def monitor_replica_status():
        while True:
            controller.policy_status_manager.maintain_life_status()
            controller.rollout_status_manager.maintain_life_status(
                controller.policy_status_manager
            )
            await asyncio.sleep(COSMOS_ROLLOUT_SCAN_INTERVAL)

    util.create_async_task(monitor_replica_status())
    yield


app = FastAPI(lifespan=lifespan)


@app.get(COSMOS_API_PANEL_SUFFIX)
async def panel():
    # HTML template with JavaScript for auto-refresh
    with open(
        os.path.join(
            os.path.dirname(__file__), "config/frontend", "dispatcher_status.html"
        ),
        "r",
        encoding="utf-8",
    ) as file:
        html = file.read()
    return HTMLResponse(html)


"""
API for replica-controller communication
"""


@app.get(COSMOS_API_STATUS_SUFFIX)
async def get_status():
    return {
        "mesh_names": MESH_NAMES,
        "policy_replicas": _serialize_replicas(controller.policy_replicas),
        "rollout_replicas": _serialize_replicas(controller.rollout_replicas),
    }


@app.get(COSMOS_API_META_SUFFIX)
async def meta():
    meta = {
        "config": controller.config,
    }
    if not controller.is_rl and controller.sft_user_dataset is not None:
        meta["sft_user_dataset"] = base64.b64encode(
            cloudpickle.dumps(controller.sft_user_dataset)
        ).decode("utf-8")
    if controller.user_data_packer is not None:
        meta["user_data_packer"] = base64.b64encode(
            cloudpickle.dumps(controller.user_data_packer)
        ).decode("utf-8")
    if controller.user_val_data_packer is not None:
        meta["user_val_data_packer"] = base64.b64encode(
            cloudpickle.dumps(controller.user_val_data_packer)
        ).decode("utf-8")
    return meta


@app.post(COSMOS_API_REGISTER_SUFFIX)
async def register(request: RegisterRequest):
    try:
        await controller.register(
            Atom.from_register_request(request),
            role=request.role,
        )
        return {"message": "Registered"}
    except Exception as e:
        import traceback

        traceback.print_exc()
        return create_error_response(constant.ErrorCode.INTERNAL_ERROR, str(e))


@app.post(COSMOS_API_UNREGISTER_SUFFIX)
async def unregister(request: UnregisterRequest):
    try:
        await controller.unregister(request.replica_name)
    except Exception as e:
        logger.error(f"[Controller] Unregister failed: {e}")
    finally:
        if (
            (
                controller.policy_status_manager.training_finished()
                or not controller.is_rl
            )
            and len(controller.policy_status_manager) == 0
            and len(controller.rollout_status_manager) == 0
        ):
            logger.info("[Controller] All replicas are finished, finalizing...")
            global server
            server.should_exit = True
        return {"message": "Unregistered"}


@app.post(COSMOS_API_SET_PROFILE_SUFFIX)
async def set_profile(request: SetProfileRequest):
    logger.info(f"[Dispatcher] set profile request: {request}")
    msg = await controller.set_profile(request)
    return msg


@app.post(COSMOS_API_SET_TRACE_PATH_SUFFIX)
async def set_trace_path(request: SetTracePathRequest):
    atom = await controller.set_trace_path(
        request.replica_name, request.trace_path, request.global_rank
    )
    if atom is not None:
        return {"message": f"Trace path set for atom: {atom}"}
    else:
        return {"message": "Ignore the trace path request!"}


@app.post(COSMOS_API_HEARTBEAT_SUFFIX)
async def heartbeat(request: HeartbeatRequest):
    # Set the replica timestamp to the current time for heartbeat
    controller.replica_heartbeat(request.replica_name)
    return {"message": "Heartbeat received"}


@app.post(COSMOS_API_POLICY_SHARD_INFOS_SUFFIX)
async def policy_shard_infos(request: Request):
    content_type = request.headers.get("Content-Type")
    if content_type != "application/msgpack":
        return create_error_response(
            constant.ErrorCode.INVALID_REQUEST,
            "Invalid Content-Type, expected application/msgpack",
        )

    raw_bytes = await request.body()
    await controller.policy_to_rollout_shard_mapper.set_shard_infos_of_policy(
        raw_bytes,
        controller.policy_status_manager.n_atoms_per_replica(),
    )
    return {"message": "Policy shard infos set"}


@app.post(COSMOS_API_ROLLOUT_SHARD_INFOS_SUFFIX)
async def rollout_shard_infos(request: Request):
    content_type = request.headers.get("Content-Type")
    if content_type != "application/msgpack":
        return create_error_response(
            constant.ErrorCode.INVALID_REQUEST,
            "Invalid Content-Type, expected application/msgpack",
        )

    raw_bytes = await request.body()
    await controller.policy_to_rollout_shard_mapper.set_shard_infos_of_rollout(
        raw_bytes,
        controller.rollout_status_manager.n_atoms_per_replica(),
    )
    return {"message": "Rollout shard infos set"}


@app.post(COSMOS_API_POLICY_SHARD_SEND_INSTS_SUFFIX)
async def policy_shard_send_insts(request: GetShardSendRecvInstsRequest):
    """
    Get the send instructions for policy.
    :return: A list of send instructions for policy.
    """
    logger.debug(
        f"[Dispatcher] Get policy shard send instructions for rank {request.rank}"
    )
    await controller.policy_to_rollout_shard_mapper.scheme_generation_done.wait()
    # Get the send instructions for policy
    send_insts = (
        await controller.policy_to_rollout_shard_mapper.get_send_insts_for_policy(
            request.rank
        )
    )
    # If the send instructions are not found, return an error response
    if send_insts is None:
        return create_error_response(
            constant.ErrorCode.INTERNAL_ERROR,
            "Policy shard send instructions not found",
        )
    logger.debug(
        f"[Dispatcher] Received policy shard send instructions for rank {request.rank}"
    )
    return Response(content=send_insts, media_type="application/msgpack")


@app.post(COSMOS_API_ROLLOUT_SHARD_RECV_INSTS_SUFFIX)
async def rollout_shard_recv_insts(request: GetShardSendRecvInstsRequest):
    """
    Get the receive instructions for rollout.
    :return: A list of receive instructions for rollout.
    """
    logger.debug(
        f"[Dispatcher] Get rollout shard receive instructions for rank {request.rank}"
    )
    # Wait for the scheme generation to be done
    await controller.policy_to_rollout_shard_mapper.scheme_generation_done.wait()
    # Get the receive instructions for rollout
    recv_insts = (
        await controller.policy_to_rollout_shard_mapper.get_recv_insts_for_rollout(
            request.rank
        )
    )
    # If the receive instructions are not found, return an error response
    if recv_insts is None:
        return create_error_response(
            constant.ErrorCode.INTERNAL_ERROR,
            "Rollout shard receive instructions not found",
        )
    logger.debug(
        f"[Dispatcher] Received rollout shard receive instructions for rank {request.rank}"
    )
    return Response(content=recv_insts, media_type="application/msgpack")


"""
NCCL Handshake API
"""


@app.post(COSMOS_API_NCCL_COMM_INITIATOR_SUFFIX)
async def comm_initiator(request: HandshakeInitiatorRequest):
    if request.handle_base64 is None or request.handle_base64 == "":
        return create_error_response(
            constant.ErrorCode.INVALID_REQUEST, "Handle is required"
        )

    await controller.update_kv_store(request.unique_pair_name, request.handle_base64)
    return {"message": "Handshake initiator received"}


@app.post(COSMOS_API_NCCL_COMM_ACCEPTOR_SUFFIX)
async def comm_acceptor(request: HandshakeAcceptorRequest):
    if request.unique_pair_name not in controller.temp_kv_store:
        return create_error_response(
            constant.ErrorCode.INTERNAL_ERROR, "Unique pair name not found"
        )
    return {"handle_base64": controller.temp_kv_store.get(request.unique_pair_name)}


@app.post(COSMOS_API_NCCL_COMM_ERROR_SUFFIX)
async def comm_error(request: NcclErrRequest):
    await controller.set_replica_ncclerror(request.replica_name, request.error)
    return {"message": "DetectTimeout received"}


@app.post(COSMOS_API_NCCL_COMM_STORE_CLEAR_SUFFIX)
async def comm_store_clear(request: NcclStoreClearRequest):
    try:
        await controller.clear_temp_kv_store(request.unique_pair_name)
    except Exception as e:
        logger.error(f"[Controller] Error clearing store: {e}")
    return {"message": "Store cleared"}


@app.get(COSMOS_API_NCCL_COMM_GET_ALL_SUFFIX)
async def comm_get_all():
    return {"comm_info": controller.temp_kv_store}


"""
Rollout API
"""


@app.get(COSMOS_API_NEXT_PROMPT_SUFFIX)
async def get_batched_prompt(n: int, validation_step: Optional[int] = None):
    prompt_id_and_payload_list, is_end = await controller.get_batched_prompt(
        n, validation_step
    )
    return {
        "prompt_id_and_payload_list": prompt_id_and_payload_list,
        "is_end": is_end,
    }


@app.post(COSMOS_API_VALIDATION_REPORT_SUFFIX)
async def validation_report(request: ValidationReportRequest):
    rollout_groups: List[RolloutGroup] = [
        RolloutGroup(
            prompt_idx=prompt_idx,
            payload=payload,
            # Only report once per replica, so is_end is always True
            is_end=True,
            completions=completions,
            reference_answer=controller.query_reference_answer(
                prompt_idx, dataset_type="val"
            ),
        )
        for prompt_idx, payload, completions in zip(
            request.prompt_idxs, request.payloads, request.completions
        )
    ]

    rollouts_list: List[List[Rollout]] = [
        rollout_group.compute_rollouts(controller.val_rl_algo)
        for rollout_group in rollout_groups
    ]
    controller.policy_status_manager.validation_report_validation_results(
        request.validation_step, rollouts_list, controller.rollout_status_manager
    )
    return {"message": "Validation rollout put"}


@app.post(COSMOS_API_ROLLOUT_SUFFIX)
async def put_rollout_group(rollout: RolloutRequest):
    try:
        if rollout.is_end:
            assert (
                len(rollout.prompt_idxs) == 0
            ), "Prompt idxs should be empty if is_end is True"
            logger.info(
                f"[Controller] Received rollout end signal from {rollout.src_replica_name}"
            )
            controller.rollout_status_manager.rollout_end(rollout.src_replica_name)
            if controller.rollout_status_manager.all_rollouts_ended():
                total_pending_rollouts = (
                    controller.policy_status_manager.total_pending_rollouts()
                )
                logger.info(
                    f"[Controller] All rollouts have ended, recompute total steps with {total_pending_rollouts} remaining rollouts..."
                )
                original_total_steps = controller.policy_status_manager.total_steps
                controller.policy_status_manager.recompute_total_steps(
                    explicit_num_remaining_samples=total_pending_rollouts
                )
                new_total_steps = controller.policy_status_manager.total_steps
                if new_total_steps > controller.policy_status_manager.current_step:
                    logger.info(
                        "[Controller] There are still remaining steps, no op required"
                    )
                    # There are still remaining steps, no op required
                    pass
                else:
                    if (
                        controller.policy_status_manager.current_step
                        == original_total_steps
                    ):
                        logger.info(
                            "[Controller] No remaining steps, policy and rollouts happen to finish at the same time"
                        )
                        # No remaining steps, policy and rollouts happen to finish at the same time
                        pass
                    else:
                        logger.info(
                            "[Controller] Clear the rollout buffer, and trigger an extra `DataFetch`"
                        )
                        # Clear the rollout buffer
                        controller.policy_status_manager.rollout_buffer.queue.clear()
                        controller.policy_status_manager.total_steps = (
                            controller.policy_status_manager.current_step + 1
                        )

                        # Trigger an extra `DataFetch & P2R/R2R`
                        controller.policy_status_manager.try_trigger_data_fetch_and_training(
                            is_fake_last_cmd=True
                        )

            return {"message": "Rollout end signal received"}

        rollout_groups: List[RolloutGroup] = [
            RolloutGroup(
                prompt_idx=prompt_idx,
                payload=payload,
                completions=completions,
                is_end=rollout.is_end,
                reference_answer=controller.query_reference_answer(prompt_idx),
            )
            for prompt_idx, payload, completions in zip(
                rollout.prompt_idxs, rollout.payloads, rollout.completions
            )
        ]

        rollouts_list: List[List[Rollout]] = [
            rollout_group.compute_rollouts(controller.rl_algo)
            for rollout_group in rollout_groups
        ]

        # Dynamic Sampling: Filter out the rollouts that the rewards are all the same
        valid_rollouts_list: List[List[Rollout]] = []
        invalid_rollouts_list: List[List[Rollout]] = []
        for rollouts_group in rollouts_list:
            if len(set([rollout.reward for rollout in rollouts_group])) > 1:
                # Preprocess the valid rollouts to find if shared prefix exists
                # If exists,
                #   - if the shared prefix hold different rewards, the prefix may lead to bias
                #   - else: do nothing
                # (shared_prefix) -> index of rollouts
                shared_prefix_groups: Dict[Tuple[int, ...], List[int]] = (
                    util.find_maximal_prefix_groups(
                        [
                            controller.tokenizer(
                                rollout.completion, add_special_tokens=False
                            ).input_ids
                            for rollout in rollouts_group
                        ],
                        N=controller.config.train.train_policy.min_filter_prefix_tokens,
                    )
                )
                for shared_prefix, rollout_indices in shared_prefix_groups.items():
                    assert (
                        len(rollout_indices) > 1
                    ), "Shared prefix group should not be empty"
                    # Check if the shared prefix holds different rewards
                    rewards = [rollouts_group[i].reward for i in rollout_indices]
                    if len(set(rewards)) > 1:
                        n_ignore_prefix_tokens = len(shared_prefix)
                        for rollout_index in rollout_indices:
                            rollouts_group[
                                rollout_index
                            ].n_ignore_prefix_tokens = n_ignore_prefix_tokens
                valid_rollouts_list.append(rollouts_group)
            else:
                # If the rewards are all the same, we need to sample one rollout from the group
                invalid_rollouts_list.append(rollouts_group)

        # Flatten the rollouts into a single list
        valid_rollouts = [
            rollout
            for rollouts_group in valid_rollouts_list
            for rollout in rollouts_group
        ]
        invalid_rollouts = [
            rollout
            for rollouts_group in invalid_rollouts_list
            for rollout in rollouts_group
        ]

        if len(valid_rollouts) > 0:
            logger.debug(
                f"[RolloutGroup] from replica: {rollout.src_replica_name} with {len(rollout.completions)} samples:"
                f"example: rollouts[0]\n{valid_rollouts[0]}"
            )

        await controller.put_rollouts(valid_rollouts, invalid_rollouts)
        return {"message": "Rollout put"}
    except Exception as e:
        import traceback

        traceback.print_exc()
        return create_error_response(constant.ErrorCode.INTERNAL_ERROR, str(e))


@app.post(COSMOS_API_POLICY_TRAIN_ACK_SUFFIX)
async def train_ack(request: TrainAckRequest):
    try:
        replicaname = request.replica_name
        step = request.weight_step
        total_steps = request.total_steps
        profile_finished = request.profile_finished
        report_data = request.report_data
        controller.policy_status_manager.train_ack(
            replicaname,
            step,
            total_steps,
            profile_finished,
            report_data,
            controller.rollout_status_manager,
        )
        return {"message": "Ack completed"}
    except Exception as e:
        import traceback

        traceback.print_exc()
        return create_error_response(constant.ErrorCode.INTERNAL_ERROR, str(e))


def _serialize_replicas(replicas: Dict[str, Replica]) -> List[Dict]:
    result = []
    for name, replica in replicas.items():
        result.append(replica.to_dict())
    return result


def main(
    dataset: Optional[Union[Dataset, Callable[[CosmosConfig], Dataset]]] = None,
    data_packer: Optional[DataPacker] = None,
    reward_fns: Optional[List[Callable]] = None,
    val_dataset: Optional[Dataset] = None,
    val_reward_fns: Optional[List[Callable]] = None,
    val_data_packer: Optional[DataPacker] = None,
    **kwargs,
):
    if kwargs:
        logger.warning(
            f"Params: {list(kwargs.keys())} are not being used in controller initialization."
        )

    # Deprecated: The following code is to ensure backward compatibility:
    # where `dispatcher` is always launched in custom script
    role = os.environ.get("COSMOS_ROLE")
    assert role in ["Policy", "Rollout", "Controller"], f"Invalid role: {role}"
    if role == "Controller":
        pass
    else:
        logger.warning(
            "Deprecated: Please update your script to use `cosmos_rl.launcher.launch()` instead of `cosmos_rl.dispatcher.run_web_panel.main`"
        )
        if role == "Policy":
            from cosmos_rl.policy.train import main as policy_main

            policy_main()
        else:
            from cosmos_rl.rollout.rollout_entrance import run_rollout

            run_rollout()
        return

    parser = argparse.ArgumentParser(
        description="Run the web panel for the dispatcher."
    )
    parser.add_argument(
        "--port", type=int, default=8000, help="Port to run the web panel on."
    )
    parser.add_argument(
        "--redis-port", type=int, default=12800, help="Port to run the web panel on."
    )
    parser.add_argument(
        "--config-file",
        type=str,
        default=None,
        required=True,
        help="Path to TOML configuration file to load.",
    )
    parser.add_argument(
        "--redis-logfile-path",
        type=str,
        default="/tmp/redis.log",
        help="The redis server log file path.",
    )
    args = parser.parse_args()

    # Load config from file if provided
    loaded_config = None
    assert os.path.exists(
        args.config_file
    ), f"Config file {args.config_file} does not exist."

    try:
        logger.info(f"Attempting to load configuration from {args.config_file}")
        with open(args.config_file, "r") as f:
            config_dict = toml.load(f)

        # Ensure CosmosConfig is available (it's imported at the top now)
        # from cosmos_rl.policy.config import Config as CosmosConfig
        # Need SFTDataConfig and GrpoConfig for from_dict

        loaded_config = CosmosConfig.from_dict(config_dict)
        # Use redis port from config if available, otherwise use arg/default
        if hasattr(loaded_config, "redis") and loaded_config.redis:
            try:
                redis_port_from_config = int(loaded_config.redis)
                args.redis_port = redis_port_from_config
                logger.info(f"Using Redis port {args.redis_port} from config file.")
            except (ValueError, TypeError):
                logger.warning(
                    f"Invalid redis port format in config file: {loaded_config.redis}. Using default/arg: {args.redis_port}"
                )

        if data_packer is not None:
            assert isinstance(
                data_packer, DataPacker
            ), "data_packer should be a DataPacker instance"
        controller.setup(
            loaded_config,
            redis_port=args.redis_port,
            redis_logfile_path=args.redis_logfile_path,
            dataset=dataset,
            reward_fns=reward_fns,
            data_packer=data_packer,
            val_dataset=val_dataset,
            val_reward_fns=val_reward_fns,
            val_data_packer=val_data_packer,
        )
        logger.info(f"Successfully loaded configuration from {args.config_file}")
    except FileNotFoundError:
        raise FileNotFoundError(f"Config file not found: {args.config_file}")
    except Exception as e:
        raise RuntimeError(
            f"Failed to load or parse config file {args.config_file}: {e}.",
            exc_info=True,
        )

    config = uvicorn.Config(
        app, host="0.0.0.0", port=util.find_available_port(args.port), access_log=False
    )
    global server
    server = uvicorn.Server(config)
    server.run()


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch.distributed as dist
import uuid
from typing import Dict, Callable, Type, Optional
import copy
import requests
import time
import atexit
from functools import partial
import threading
from urllib.parse import urljoin
from cosmos_rl.utils.redis_stream import RedisStreamHandler
from cosmos_rl.utils.network_util import make_request_with_retry, get_local_ip
from cosmos_rl.dispatcher.command import CommandRegistry, Command
from cosmos_rl.dispatcher.data.packer import DataPacker, DecoderOnlyLLMDataPacker
from cosmos_rl.utils.logging import logger
import cosmos_rl.utils.constant as constant
import cosmos_rl.utils.distributed as dist_utils
from cosmos_rl.dispatcher.protocol import MESH_NAMES
import cosmos_rl.utils.util as util
from cosmos_rl.utils.api_suffix import (
    COSMOS_API_REGISTER_SUFFIX,
    COSMOS_API_UNREGISTER_SUFFIX,
    COSMOS_API_HEARTBEAT_SUFFIX,
)
import base64
import cloudpickle
from transformers import AutoConfig


class CommMixin:
    policy_command_handler_registry = CommandRegistry()
    rollout_command_handler_registry = CommandRegistry()

    @classmethod
    def register_policy_command_handler(cls, command_type: Type[Command]):
        def decorator(func):
            cls.policy_command_handler_registry.register(command_type, func)
            return func

        return decorator

    @classmethod
    def register_rollout_command_handler(cls, command_type: Type[Command]):
        def decorator(func):
            cls.rollout_command_handler_registry.register(command_type, func)
            return func

        return decorator

    @classmethod
    def get_policy_command_handler(
        cls, command_type: Type[Command]
    ) -> Optional[Callable]:
        return cls.policy_command_handler_registry.get_command_handler(command_type)

    @classmethod
    def get_rollout_command_handler(
        cls, command_type: Type[Command]
    ) -> Optional[Callable]:
        return cls.rollout_command_handler_registry.get_command_handler(command_type)

    def init_comm(self):
        self.replica_name = str(dist_utils.broadcast_object_cpu(uuid.uuid4()))
        logger.info(
            f"{self.role} Replica started at global rank {self.global_rank}, with replica name: {self.replica_name}"
        )
        # Fetch metadata from the controller
        # Get list of remote IPs and port
        self.remote_ips, self.remote_port, metadata = (
            dist_utils.get_controller_metadata()
        )
        # `sft_user_dataset` is only used in SFT mode when the user provides a dataset
        self.sft_user_dataset = None
        sft_user_dataset = metadata.get("sft_user_dataset", None)
        if sft_user_dataset:
            sft_user_dataset = base64.b64decode(sft_user_dataset)
            sft_user_dataset = cloudpickle.loads(sft_user_dataset)
            if hasattr(sft_user_dataset, "setup"):
                sft_user_dataset.setup(self.config, self.tokenizer)
            self.sft_user_dataset = sft_user_dataset

        hf_config = util.retry(AutoConfig.from_pretrained)(
            self.config.policy.model_name_or_path, trust_remote_code=True
        )

        user_data_packer = metadata.get("user_data_packer", None)
        if user_data_packer:
            user_data_packer = base64.b64decode(user_data_packer)
            user_data_packer = cloudpickle.loads(user_data_packer)
            self.data_packer = user_data_packer
            logger.info(f"Using user-provided data packer: {self.data_packer}")
        else:
            try:
                self.data_packer = DataPacker.get_default_data_packer(
                    hf_config.model_type
                )
                logger.info(f"Using default data packer: {self.data_packer}")
            except ValueError:
                logger.warning(
                    f"No default data packer found for {hf_config.model_type}, using DecoderOnlyLLMDataPacker as default"
                )
                self.data_packer = DecoderOnlyLLMDataPacker()
        self.data_packer.setup(self.config, self.tokenizer)

        user_val_data_packer = metadata.get("user_val_data_packer", None)
        if user_val_data_packer:
            user_val_data_packer = base64.b64decode(user_val_data_packer)
            user_val_data_packer = cloudpickle.loads(user_val_data_packer)
            self.val_data_packer = user_val_data_packer
            logger.info(
                f"Using user-provided validation data packer: {self.val_data_packer}"
            )
        else:
            try:
                self.val_data_packer = DataPacker.get_default_data_packer(
                    hf_config.model_type
                )
                logger.info(
                    f"Using default validation data packer: {self.val_data_packer}"
                )
            except ValueError:
                logger.warning(
                    f"No default validation data packer found for {hf_config.model_type}, using DecoderOnlyLLMDataPacker as default"
                )
                self.val_data_packer = DecoderOnlyLLMDataPacker()
        self.val_data_packer.setup(self.config, self.tokenizer)

        self.remote_hosts = [
            f"http://{remote_ip}:{self.remote_port}" for remote_ip in self.remote_ips
        ]
        self.register_to_controller()

    def register_to_controller(self):
        if hasattr(self, "_is_registered"):
            return

        target_mesh_names = copy.deepcopy(MESH_NAMES)
        ranks = []
        group_size = []
        for mesh_name in MESH_NAMES:
            if (
                self.parallel_dims.mesh.mesh_dim_names
                and mesh_name in self.parallel_dims.mesh.mesh_dim_names
            ):
                ranks.append(self.parallel_dims.mesh[mesh_name].get_local_rank())
                group_size.append(self.parallel_dims.mesh[mesh_name].size())
            else:
                ranks.append(0)
                group_size.append(1)

        try:
            host_info_tuple = get_local_ip()
            if host_info_tuple is None:
                raise Exception("Failed to get local IP address")
            host_ip, host_name = host_info_tuple

            payload = {
                "replica_name": self.replica_name,
                "role": self.role,
                "mesh_names": target_mesh_names,
                "ranks": ranks,
                "group_size": group_size,
                "global_rank": self.global_rank,
                "host_ip": host_ip,
                "host_name": host_name,
            }
            make_request_with_retry(
                partial(
                    requests.post,
                    json=payload,
                ),
                self.get_alternative_urls(COSMOS_API_REGISTER_SUFFIX),
                max_retries=constant.COSMOS_HTTP_RETRY_CONFIG.max_retries,
            )
        except Exception as e:
            logger.error(f"Failed to register to controller: {e}")
            raise e

        dist.barrier()  # wait all the atoms registered.

        self.shutdown_signal = threading.Event()

        if self.global_rank == 0:
            logger.info(
                f"{self.role} Replica {self.replica_name} registered to controller"
            )
            # Start the thread with daemon=True, so it will exit when the main program exits.
            thread = threading.Thread(
                target=self.heartbeat_trigger,
                args=(self.shutdown_signal,),
                daemon=True,
            )
            thread.start()
            self.heartbeat_thread = thread
        else:
            self.heartbeat_thread = None

        self._is_registered = True
        atexit.register(self.unregister_from_controller)

    def unregister_from_controller(self):
        if not hasattr(self, "_is_registered"):
            return
        elif hasattr(self, "_is_unregistered"):
            return
        else:
            self._is_unregistered = True
        self._is_registered = False
        # let only rank == 0 send the unregister request
        if self.global_rank == 0:
            try:
                make_request_with_retry(
                    partial(
                        requests.post,
                        json={"replica_name": self.replica_name},
                    ),
                    self.get_alternative_urls(COSMOS_API_UNREGISTER_SUFFIX),
                    max_retries=constant.COSMOS_HTTP_RETRY_CONFIG.max_retries,
                )
            except Exception as e:
                logger.error(f"Failed to unregister to controller: {e}")

    def get_group_unique_key(self, replica_name_to_rank: Dict[str, int]):
        return (
            "_".join(
                [
                    k
                    for k, _ in sorted(
                        replica_name_to_rank.items(), key=lambda item: item[1]
                    )
                ]
            )
            + "_"
            + str(self.global_rank)
        )

    def init_redis(self):
        # For command fetch via redis connection
        self.redis_controller = RedisStreamHandler(
            ips=self.remote_ips, port=int(self.config.redis)
        )
        logger.debug(
            f"[{self.role}] Init redis at {self.remote_ips}:{self.redis_controller.port}"
        )

    def heartbeat_trigger(self, shutdown_signal: threading.Event):
        while True:
            try:
                make_request_with_retry(
                    partial(
                        requests.post,
                        json={
                            "replica_name": self.replica_name,
                        },
                    ),
                    self.get_alternative_urls(COSMOS_API_HEARTBEAT_SUFFIX),
                    max_retries=constant.COSMOS_HTTP_RETRY_CONFIG.max_retries,
                )
            except Exception as e:
                logger.error(f"Failed to send heartbeat to controller: {e}")

            # If the heartbeat interval is greater than 1, we need to check the shutdown signal every second
            # for faster shutdown check
            if constant.COSMOS_HEARTBEAT_SEND_INTERVAL > 1:
                early_break = False
                for _ in range(int(constant.COSMOS_HEARTBEAT_SEND_INTERVAL)):
                    if shutdown_signal.is_set():
                        early_break = True
                        break
                    else:
                        time.sleep(1)
                if early_break:
                    break
            else:
                time.sleep(constant.COSMOS_HEARTBEAT_SEND_INTERVAL)
                if shutdown_signal.is_set():
                    break

    def get_alternative_urls(self, suffix: str):
        # Get the alternative URLs for the given suffix
        urls = []
        for remote_host in self.remote_hosts:
            urls.append(urljoin(remote_host, suffix))
        return urls
